from birder.net.mim.crossmae import CrossMAE
from birder.net.mim.fcmae import FCMAE
from birder.net.mim.mae_hiera import MAE_Hiera
from birder.net.mim.mae_vit import MAE_ViT
from birder.net.mim.simmim import SimMIM

__all__ = [
    "CrossMAE",
    "FCMAE",
    "MAE_Hiera",
    "MAE_ViT",
    "SimMIM",
]



import logging
import os
import warnings
from pathlib import Path
from types import ModuleType
from typing import Optional

import torch
from torch.utils.cpp_extension import load

import birder

logger = logging.getLogger(__name__)


_CACHED_KERNELS: dict[str, ModuleType] = {}


def load_msda() -> Optional[ModuleType]:
    name = "msda"
    if torch.cuda.is_available() is False or os.environ.get("DISABLE_CUSTOM_KERNELS", "0") == "1":
        return None

    if name in _CACHED_KERNELS:
        return _CACHED_KERNELS[name]

    # Adapted from:
    # https://github.com/huggingface/transformers/blob/main/src/transformers/models/deformable_detr/load_custom.py
    root = Path(birder.__file__).resolve().parent.joinpath("kernels/deformable_detr")
    src_files = [
        root.joinpath("vision.cpp"),
        root.joinpath("cpu/ms_deform_attn_cpu.cpp"),
        root.joinpath("cuda/ms_deform_attn_cuda.cu"),
    ]

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        msda: Optional[ModuleType] = load(
            "MultiScaleDeformableAttention",
            src_files,
            with_cuda=True,
            extra_include_paths=[str(root)],
            extra_cflags=["-DWITH_CUDA=1"],
            extra_cuda_cflags=[
                "-DCUDA_HAS_FP16=1",
                "-D__CUDA_NO_HALF_OPERATORS__",
                "-D__CUDA_NO_HALF_CONVERSIONS__",
                "-D__CUDA_NO_HALF2_OPERATORS__",
            ],
        )

    if msda is not None:
        logger.info("MSDA custom kernel loaded")
        _CACHED_KERNELS[name] = msda
    else:
        logger.debug("MSDA custom kernel NOT loaded")

    return msda


def load_swattention() -> Optional[ModuleType]:
    name = "swattention"
    if torch.cuda.is_available() is False or os.environ.get("DISABLE_CUSTOM_KERNELS", "0") == "1":
        return None

    if name in _CACHED_KERNELS:
        return _CACHED_KERNELS[name]

    root = Path(birder.__file__).resolve().parent.joinpath("kernels/transnext")
    src_files = [
        root.joinpath("swattention.cpp"),
        root.joinpath("av_bw_kernel.cu"),
        root.joinpath("av_fw_kernel.cu"),
        root.joinpath("qk_bw_kernel.cu"),
        root.joinpath("qk_fw_kernel.cu"),
        root.joinpath("qk_rpb_bw_kernel.cu"),
        root.joinpath("qk_rpb_fw_kernel.cu"),
    ]

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        swattention: Optional[ModuleType] = load(
            "swattention",
            src_files,
            with_cuda=True,
            extra_cflags=["-DWITH_CUDA=1"],
            extra_cuda_cflags=[
                "-DCUDA_HAS_FP16=1",
                "-D__CUDA_NO_HALF_OPERATORS__",
                "-D__CUDA_NO_HALF_CONVERSIONS__",
                "-D__CUDA_NO_HALF2_OPERATORS__",
            ],
        )

    if swattention is not None:
        logger.info("swattention custom kernel loaded")
        _CACHED_KERNELS[name] = swattention
    else:
        logger.debug("swattention custom kernel NOT loaded")

    return swattention


def load_soft_nms() -> Optional[ModuleType]:
    name = "soft_nms"
    if os.environ.get("DISABLE_CUSTOM_KERNELS", "0") == "1":
        return None

    if name in _CACHED_KERNELS:
        return _CACHED_KERNELS[name]

    root = Path(birder.__file__).resolve().parent.joinpath("kernels/soft_nms")
    src_files = [
        root.joinpath("op.cpp"),
        root.joinpath("soft_nms.cpp"),
    ]

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        soft_nms: Optional[ModuleType] = load(
            "soft_nms",
            src_files,
            with_cuda=True,
            extra_cflags=["-DWITH_CUDA=1"],
            extra_cuda_cflags=[
                "-DCUDA_HAS_FP16=1",
                "-D__CUDA_NO_HALF_OPERATORS__",
                "-D__CUDA_NO_HALF_CONVERSIONS__",
                "-D__CUDA_NO_HALF2_OPERATORS__",
            ],
        )

    if soft_nms is not None:
        logger.info("soft_nms custom kernel loaded")
        _CACHED_KERNELS[name] = soft_nms
    else:
        logger.debug("soft_nms custom kernel NOT loaded")

    return soft_nms



"""
LeViT, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/levit.py

Paper "LeViT: a Vision Transformer in ConvNet's Clothing for Faster Inference", https://arxiv.org/abs/2104.01136

Changes from original:
* Removed attention bias cache
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.base import interpolate_attention_bias
from birder.net.vit import PatchEmbed


class LinearNorm(nn.Module):
    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias=False)
        self.bn = nn.BatchNorm1d(out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        return self.bn(x.flatten(0, 1)).reshape_as(x)


class Subsample(nn.Module):
    def __init__(self, stride: int, resolution: tuple[int, int]) -> None:
        super().__init__()
        self.stride = stride
        self.resolution = resolution

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, C) = x.shape
        x = x.view(B, self.resolution[0], self.resolution[1], C)
        x = x[:, :: self.stride, :: self.stride]
        return x.reshape(B, -1, C)


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        resolution: tuple[int, int],
        act_layer: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = key_dim**-0.5
        self.key_dim = key_dim
        self.val_dim = int(attn_ratio * key_dim)
        self.val_attn_dim = int(attn_ratio * key_dim) * num_heads

        key_attn_dim = key_dim * num_heads
        self.qkv = LinearNorm(dim, self.val_attn_dim + key_attn_dim * 2)
        self.proj = nn.Sequential(
            act_layer(),
            LinearNorm(self.val_attn_dim, dim),
        )

        self.attention_biases = nn.Parameter(torch.zeros(self.num_heads, resolution[0] * resolution[1]))
        pos = torch.stack(
            torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
        ).flatten(1)
        rel_pos = (pos[..., :, None] - pos[..., None, :]).abs()
        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
        self.attention_bias_idxs = nn.Buffer(rel_pos, persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, _) = x.shape
        q, k, v = self.qkv(x).view(B, N, self.num_heads, -1).split([self.key_dim, self.key_dim, self.val_dim], dim=3)
        q = q.permute(0, 2, 1, 3)
        k = k.permute(0, 2, 3, 1)
        v = v.permute(0, 2, 1, 3)

        attn = q @ k * self.scale + self.attention_biases[:, self.attention_bias_idxs]
        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, N, self.val_attn_dim)
        x = self.proj(x)

        return x


class AttentionSubsample(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        stride: int,
        resolution: tuple[int, int],
        act_layer: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        self.stride = stride
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.val_dim = int(attn_ratio * key_dim)
        self.val_attn_dim = self.val_dim * self.num_heads
        self.scale = key_dim**-0.5

        key_attn_dim = key_dim * num_heads
        self.kv = LinearNorm(in_dim, self.val_attn_dim + key_attn_dim)
        self.q = nn.Sequential(
            Subsample(stride, resolution),
            LinearNorm(in_dim, key_attn_dim),
        )
        self.proj = nn.Sequential(
            act_layer(),
            LinearNorm(self.val_attn_dim, out_dim),
        )

        self.attention_biases = nn.Parameter(torch.zeros(self.num_heads, resolution[0] * resolution[1]))
        k_pos = torch.stack(
            torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
        ).flatten(1)
        q_pos = torch.stack(
            torch.meshgrid(
                torch.arange(0, resolution[0], step=stride), torch.arange(0, resolution[1], step=stride), indexing="ij"
            )
        ).flatten(1)
        rel_pos = (q_pos[..., :, None] - k_pos[..., None, :]).abs()
        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
        self.attention_bias_idxs = nn.Buffer(rel_pos, persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, _) = x.shape
        k, v = self.kv(x).view(B, N, self.num_heads, -1).split([self.key_dim, self.val_dim], dim=3)
        k = k.permute(0, 2, 3, 1)  # BHCN
        v = v.permute(0, 2, 1, 3)  # BHNC
        q = self.q(x).view(B, -1, self.num_heads, self.key_dim).permute(0, 2, 1, 3)

        attn = q @ k * self.scale + self.attention_biases[:, self.attention_bias_idxs]
        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, -1, self.val_attn_dim)
        x = self.proj(x)

        return x


class LeVitMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        act_layer: Callable[..., nn.Module],
    ) -> None:
        super().__init__()
        self.ln1 = LinearNorm(in_features, hidden_features)
        self.act = act_layer()
        self.ln2 = LinearNorm(hidden_features, in_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.ln1(x)
        x = self.act(x)
        x = self.ln2(x)

        return x


class LeVitSubsample(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        mlp_ratio: float,
        act_layer: Callable[..., nn.Module],
        resolution: tuple[int, int],
        drop_path: float,
    ) -> None:
        super().__init__()
        self.attn_downsample = AttentionSubsample(
            in_dim=in_dim,
            out_dim=out_dim,
            key_dim=key_dim,
            num_heads=num_heads,
            attn_ratio=attn_ratio,
            stride=2,
            resolution=resolution,
            act_layer=act_layer,
        )

        self.mlp = LeVitMLP(out_dim, int(out_dim * mlp_ratio), act_layer=act_layer)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.attn_downsample(x)
        x = x + self.drop_path(self.mlp(x))
        return x


class LeVitBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        key_dim: int,
        num_heads: int,
        attn_ratio: float,
        mlp_ratio: float,
        resolution: tuple[int, int],
        act_layer: Callable[..., nn.Module],
        drop_path: float,
    ) -> None:
        super().__init__()
        self.attn = Attention(
            dim=dim,
            key_dim=key_dim,
            num_heads=num_heads,
            attn_ratio=attn_ratio,
            resolution=resolution,
            act_layer=act_layer,
        )
        self.drop_path1 = StochasticDepth(drop_path, mode="row")

        self.mlp = LeVitMLP(dim, int(dim * mlp_ratio), act_layer=act_layer)
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path1(self.attn(x))
        x = x + self.drop_path2(self.mlp(x))
        return x


class LevitStage(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        key_dim: int,
        depth: int,
        num_heads: int,
        attn_ratio: float,
        mlp_ratio: float,
        act_layer: Callable[..., nn.Module],
        resolution: tuple[int, int],
        downsample: bool,
        drop_path: float,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = LeVitSubsample(
                in_dim,
                out_dim,
                key_dim=key_dim,
                num_heads=in_dim // key_dim,
                attn_ratio=4.0,
                mlp_ratio=2.0,
                act_layer=act_layer,
                resolution=resolution,
                drop_path=drop_path,
            )
            resolution = ((resolution[0] - 1) // 2 + 1, (resolution[1] - 1) // 2 + 1)
        else:
            assert in_dim == out_dim
            self.downsample = nn.Identity()

        blocks = []
        for _ in range(depth):
            blocks.append(
                LeVitBlock(
                    out_dim,
                    key_dim,
                    num_heads=num_heads,
                    attn_ratio=attn_ratio,
                    mlp_ratio=mlp_ratio,
                    resolution=resolution,
                    act_layer=act_layer,
                    drop_path=drop_path,
                )
            )
        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)
        return x


class LeViT(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        embed_dim: list[int] = self.config["embed_dim"]
        key_dim: int = self.config["key_dim"]
        num_heads: list[int] = self.config["num_heads"]
        depths: list[int] = self.config["depths"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                embed_dim[0] // 8,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.Hardswish,
            ),
            Conv2dNormActivation(
                embed_dim[0] // 8,
                embed_dim[0] // 4,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.Hardswish,
            ),
            Conv2dNormActivation(
                embed_dim[0] // 4,
                embed_dim[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.Hardswish,
            ),
            Conv2dNormActivation(
                embed_dim[0] // 2,
                embed_dim[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.Hardswish,
            ),
            PatchEmbed(),
        )

        resolution = (self.size[0] // 16, self.size[1] // 16)
        in_dim = embed_dim[0]
        num_stages = len(depths)
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        for i in range(num_stages):
            if i == 0:
                stage_stride = 1
            else:
                stage_stride = 2

            stages[f"stage{i+1}"] = LevitStage(
                in_dim,
                embed_dim[i],
                key_dim=key_dim,
                depth=depths[i],
                num_heads=num_heads[i],
                attn_ratio=2.0,
                mlp_ratio=2.0,
                act_layer=nn.Hardswish,
                resolution=resolution,
                downsample=stage_stride == 2,
                drop_path=drop_path_rate,  # No stochastic depth decay rule, simple float
            )
            resolution = ((resolution[0] - 1) // stage_stride + 1, (resolution[1] - 1) // stage_stride + 1)
            in_dim = embed_dim[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Permute([0, 2, 1]),
            nn.BatchNorm1d(embed_dim[-1]),
            nn.AdaptiveAvgPool1d(output_size=1),
            nn.Flatten(1),
        )
        self.embedding_size = embed_dim[-1]
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = self.classifier(x)
        x_dist = self.dist_classifier(x)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            # Classifier "token" as an average of both tokens (during normal training or inference)
            x = (x_cls + x_dist) / 2

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        old_resolution = (old_size[0] // 16, old_size[1] // 16)
        resolution = (new_size[0] // 16, new_size[1] // 16)
        for stage in self.body:
            if isinstance(stage, LevitStage):
                for m in stage.modules():
                    if isinstance(m, AttentionSubsample):
                        # Update Subsample resolution
                        m.q[0].resolution = resolution

                        # Interpolate attention biases
                        m.attention_biases = nn.Parameter(
                            interpolate_attention_bias(m.attention_biases, old_resolution, resolution)
                        )

                        # Rebuild attention bias indices
                        k_pos = torch.stack(
                            torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
                        ).flatten(1)
                        q_pos = torch.stack(
                            torch.meshgrid(
                                torch.arange(0, resolution[0], step=m.stride),
                                torch.arange(0, resolution[1], step=m.stride),
                                indexing="ij",
                            )
                        ).flatten(1)
                        rel_pos = (q_pos[..., :, None] - k_pos[..., None, :]).abs()
                        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
                        m.attention_bias_idxs = nn.Buffer(rel_pos, persistent=False)

                        old_resolution = ((old_resolution[0] - 1) // 2 + 1, (old_resolution[1] - 1) // 2 + 1)
                        resolution = ((resolution[0] - 1) // 2 + 1, (resolution[1] - 1) // 2 + 1)

                    elif isinstance(m, Attention):
                        # Interpolate attention biases
                        m.attention_biases = nn.Parameter(
                            interpolate_attention_bias(m.attention_biases, old_resolution, resolution)
                        )

                        # Rebuild attention bias indices
                        pos = torch.stack(
                            torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
                        ).flatten(1)
                        rel_pos = (pos[..., :, None] - pos[..., None, :]).abs()
                        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
                        m.attention_bias_idxs = nn.Buffer(rel_pos, persistent=False)


registry.register_model_config(
    "levit_128s",
    LeViT,
    config={
        "embed_dim": [128, 256, 384],
        "key_dim": 16,
        "num_heads": [4, 6, 8],
        "depths": [2, 3, 4],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "levit_128",
    LeViT,
    config={
        "embed_dim": [128, 256, 384],
        "key_dim": 16,
        "num_heads": [4, 8, 12],
        "depths": [4, 4, 4],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "levit_192",
    LeViT,
    config={
        "embed_dim": [192, 288, 384],
        "key_dim": 32,
        "num_heads": [3, 5, 6],
        "depths": [4, 4, 4],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "levit_256",
    LeViT,
    config={
        "embed_dim": [256, 384, 512],
        "key_dim": 32,
        "num_heads": [4, 6, 8],
        "depths": [4, 4, 4],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "levit_384",
    LeViT,
    config={
        "embed_dim": [384, 512, 768],
        "key_dim": 32,
        "num_heads": [6, 9, 12],
        "depths": [4, 4, 4],
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "levit_128s_il-common",
    {
        "description": "LeViT 128s model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 28.1,
                "sha256": "f70629adc76d029445f6eb70786075e71b9e7d99029186fefc4e1fb966aaf743",
            }
        },
        "net": {"network": "levit_128s", "tag": "il-common"},
    },
)
registry.register_weights(
    "levit_128_il-common",
    {
        "description": "LeViT 128 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 33.6,
                "sha256": "877006385c7bb9c9a48bb602803c10075a752f522941f085b241e5dfcdc77f9c",
            }
        },
        "net": {"network": "levit_128", "tag": "il-common"},
    },
)



import logging
import unittest

import torch

from birder import layers

logging.disable(logging.CRITICAL)


class TestLayers(unittest.TestCase):
    def test_activations(self) -> None:
        quick_gelu = layers.QuickGELU()
        out = quick_gelu(torch.rand(2, 8))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 8))

    def test_attention_pool(self) -> None:
        attention_pool = layers.MultiHeadAttentionPool(32, 2, 64, True)
        out = attention_pool(torch.rand(2, 8, 32))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 1, 32))

    def test_ffn(self) -> None:
        swiglu_ffn = layers.FFN(8, 16)
        out = swiglu_ffn(torch.rand(2, 8))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 8))

    def test_swiglu_ffn(self) -> None:
        swiglu_ffn = layers.SwiGLU_FFN(8, 16)
        out = swiglu_ffn(torch.rand(2, 8))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 8))

    def test_gem(self) -> None:
        fixed_gem = layers.FixedGeMPool2d(3)
        out = fixed_gem(torch.rand(2, 8, 16, 16))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 8))

        gem = layers.GeMPool2d(3)
        out = gem(torch.rand(2, 8, 16, 16))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 8))

    def test_layer_norm(self) -> None:
        ln = layers.LayerNorm2d(16)
        out = ln(torch.rand(1, 16, 64, 64))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (1, 16, 64, 64))

    def test_layer_scale(self) -> None:
        ls = layers.LayerScale(16, 1e-5)

        # 1D
        out = ls(torch.rand(1, 64, 16))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (1, 64, 16))

        # 2D channels last
        out = ls(torch.rand(1, 64, 64, 16))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (1, 64, 64, 16))

    def test_layer_scale2d(self) -> None:
        ls = layers.LayerScale2d(16, 1e-5)
        out = ls(torch.rand(2, 16, 64, 64))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.size(), (2, 16, 64, 64))



"""
BYOL, adapted from
https://github.com/lucidrains/byol-pytorch/blob/master/byol_pytorch/byol_pytorch.py

Paper "Bootstrap your own latent: A new approach to self-supervised Learning",
https://arxiv.org/abs/2006.07733
"""

# Reference license: MIT

import copy
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


def loss_fn(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    x = F.normalize(x, dim=-1)
    y = F.normalize(y, dim=-1)
    return 2 - 2 * (x * y).sum(dim=-1)


class MLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, out_features: int) -> None:
        super().__init__()
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.norm = nn.BatchNorm1d(hidden_features)
        self.act = nn.ReLU(inplace=True)
        self.fc2 = nn.Linear(hidden_features, out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.fc2(x)

        return x


class BYOLEncoder(nn.Module):
    def __init__(self, backbone: BaseNet, projection_size: int, projection_hidden_size: int):
        super().__init__()
        self.backbone = backbone
        self.projector = MLP(backbone.embedding_size, projection_hidden_size, projection_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.backbone.embedding(x)
        return self.projector(x)


class BYOL(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"

        projection_size: int = self.config["projection_size"]
        projection_hidden_size: int = self.config["projection_hidden_size"]

        target_encoder_backbone = copy.deepcopy(self.backbone)

        self.online_encoder = BYOLEncoder(self.backbone, projection_size, projection_hidden_size)
        self.target_encoder = BYOLEncoder(target_encoder_backbone, projection_size, projection_hidden_size)
        self.online_predictor = MLP(projection_size, projection_hidden_size, projection_size)

        # Weights initialization
        self.target_encoder.load_state_dict(self.online_encoder.state_dict())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        projection = self.online_encoder(x)
        online_predictions = self.online_predictor(projection)
        (online_pred_one, online_pred_two) = online_predictions.chunk(2, dim=0)

        target_projections = self.target_encoder(x)
        (target_proj_one, target_proj_two) = target_projections.chunk(2, dim=0)

        loss_one = loss_fn(online_pred_one, target_proj_two.detach())
        loss_two = loss_fn(online_pred_two, target_proj_one.detach())
        loss = loss_one + loss_two

        return loss.mean()



"""
ViTDet (Vision Transformer for Object Detection).

Paper "Exploring Plain Vision Transformer Backbones for Object Detection",
https://arxiv.org/abs/2203.16527

Changes from original:
* Uses Faster R-CNN instead of Mask R-CNN, performing object detection without mask prediction
* Intended to use ViT-SAM (Vision Transformer with spatial modifications) as the backbone
"""

from typing import Any
from typing import Optional

from torch import nn
from torchvision.ops import MultiScaleRoIAlign
from torchvision.ops.feature_pyramid_network import LastLevelMaxPool

from birder.net.base import DetectorBackbone
from birder.net.detection.base import AnchorGenerator
from birder.net.detection.base import BackboneWithSimpleFPN
from birder.net.detection.faster_rcnn import Faster_RCNN
from birder.net.detection.faster_rcnn import FastRCNNConvFCHead
from birder.net.detection.faster_rcnn import FastRCNNPredictor
from birder.net.detection.faster_rcnn import RegionProposalNetwork
from birder.net.detection.faster_rcnn import RoIHeads
from birder.net.detection.faster_rcnn import RPNHead


class ViTDet(Faster_RCNN):
    default_size = (640, 640)
    auto_register = True

    # pylint: disable=too-many-locals
    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is None, "config not supported"

        fpn_width = 256
        rpn_pre_nms_top_n_train = 2000
        rpn_pre_nms_top_n_test = 1000
        rpn_post_nms_top_n_train = 2000
        rpn_post_nms_top_n_test = 1000
        rpn_fg_iou_thresh = 0.7
        rpn_bg_iou_thresh = 0.3
        rpn_batch_size_per_image = 256
        rpn_positive_fraction = 0.5
        rpn_nms_thresh = 0.7
        rpn_score_thresh = 0.0
        box_fg_iou_thresh = 0.5
        box_bg_iou_thresh = 0.5
        box_batch_size_per_image = 512
        box_positive_fraction = 0.25
        bbox_reg_weights = (10.0, 10.0, 5.0, 5.0)
        box_score_thresh = 0.05
        box_nms_thresh = 0.5
        box_detections_per_img = 100
        canonical_scale = 224

        self.backbone_with_fpn = BackboneWithSimpleFPN(self.backbone, fpn_width, extra_blocks=LastLevelMaxPool())

        anchor_sizes = [[32], [64], [128], [256], [512]]
        aspect_ratios = [[0.5, 1.0, 2.0]] * len(anchor_sizes)
        rpn_anchor_generator = AnchorGenerator(anchor_sizes, aspect_ratios)
        rpn_head = RPNHead(
            self.backbone_with_fpn.out_channels, rpn_anchor_generator.num_anchors_per_location()[0], conv_depth=2
        )

        rpn_pre_nms_top_n = {"training": rpn_pre_nms_top_n_train, "testing": rpn_pre_nms_top_n_test}
        rpn_post_nms_top_n = {"training": rpn_post_nms_top_n_train, "testing": rpn_post_nms_top_n_test}

        self.rpn = RegionProposalNetwork(
            rpn_anchor_generator,
            rpn_head,
            rpn_fg_iou_thresh,
            rpn_bg_iou_thresh,
            rpn_batch_size_per_image,
            rpn_positive_fraction,
            rpn_pre_nms_top_n,
            rpn_post_nms_top_n,
            rpn_nms_thresh,
            score_thresh=rpn_score_thresh,
        )
        box_roi_pool = MultiScaleRoIAlign(
            featmap_names=["stage1", "stage2", "stage3", "stage4"],  # Generated by the simple FPN
            output_size=canonical_scale // 32,
            sampling_ratio=2,
            canonical_scale=canonical_scale,
        )
        box_head = FastRCNNConvFCHead(
            (self.backbone_with_fpn.out_channels, canonical_scale // 32, canonical_scale // 32),
            [256, 256, 256, 256],
            [1024],
            norm_layer=nn.BatchNorm2d,
        )

        self.representation_size = 1024
        box_predictor = FastRCNNPredictor(self.representation_size, self.num_classes)

        self.roi_heads = RoIHeads(
            box_roi_pool,
            box_head,
            box_predictor,
            box_fg_iou_thresh,
            box_bg_iou_thresh,
            box_batch_size_per_image,
            box_positive_fraction,
            bbox_reg_weights,
            box_score_thresh,
            box_nms_thresh,
            box_detections_per_img,
            export_mode=self.export_mode,
        )



"""
Conv2Former, adapted from
https://github.com/HVision-NKU/Conv2Former/blob/main/models/conv2former.py

Paper "Conv2Former: A Simple Transformer-Style ConvNet for Visual Recognition", https://arxiv.org/abs/2211.11943
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.hornet import ChannelsFirstLayerNorm


class MLP(nn.Module):
    def __init__(self, dim: int, mlp_ratio: float) -> None:
        super().__init__()
        expanded_channels = int(dim * mlp_ratio)
        self.norm = ChannelsFirstLayerNorm(dim, eps=1e-6)
        self.fc1 = nn.Conv2d(dim, expanded_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.pos = nn.Conv2d(
            expanded_channels,
            expanded_channels,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            groups=expanded_channels,
        )
        self.fc2 = nn.Conv2d(expanded_channels, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(x)
        x = self.fc1(x)
        x = self.act(x)
        x = x + self.act(self.pos(x))
        x = self.fc2(x)

        return x


class SpatialAttention(nn.Module):
    def __init__(self, dim: int, kernel_size: tuple[int, int]) -> None:
        super().__init__()
        self.norm = ChannelsFirstLayerNorm(dim, eps=1e-6)
        self.attn = nn.Sequential(
            nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
            nn.GELU(),
            nn.Conv2d(
                dim,
                dim,
                kernel_size=kernel_size,
                stride=(1, 1),
                padding=(kernel_size[0] // 2, kernel_size[1] // 2),
                groups=dim,
            ),
        )
        self.v = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.proj = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(x)
        x = self.attn(x) * self.v(x)
        x = self.proj(x)

        return x


class Conv2FormerBlock(nn.Module):
    def __init__(self, dim: int, kernel_size: tuple[int, int], mlp_ratio: float, drop_path: float) -> None:
        super().__init__()
        self.attn = SpatialAttention(dim, kernel_size)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.mlp = MLP(dim, mlp_ratio)

        layer_scale_init_value = 1e-6
        self.layer_scale_1 = nn.Parameter(layer_scale_init_value * torch.ones((1, dim, 1, 1)), requires_grad=True)
        self.layer_scale_2 = nn.Parameter(layer_scale_init_value * torch.ones((1, dim, 1, 1)), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1 * self.attn(x))
        x = x + self.drop_path(self.layer_scale_2 * self.mlp(x))

        return x


class Conv2FormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        kernel_size: tuple[int, int],
        mlp_ratio: float,
        drop_path: list[float],
        depth: int,
        downsample: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = nn.Sequential(
                ChannelsFirstLayerNorm(dim, eps=1e-6),
                nn.Conv2d(dim, out_dim, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
            )
        else:
            assert dim == out_dim
            self.downsample = nn.Identity()

        layers = []
        for i in range(depth):
            layers.append(
                Conv2FormerBlock(out_dim, kernel_size=kernel_size, mlp_ratio=mlp_ratio, drop_path=drop_path[i])
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class Conv2Former(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.blocks.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_ratios = [4.0, 4.0, 4.0, 4.0]
        kernel_size: tuple[int, int] = self.config["kernel_size"]
        dims: list[int] = self.config["dims"]
        depths: list[int] = self.config["depths"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, dims[0] // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False),
            nn.GELU(),
            nn.BatchNorm2d(dims[0] // 2),
            nn.Conv2d(dims[0] // 2, dims[0] // 2, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
            nn.GELU(),
            nn.BatchNorm2d(dims[0] // 2),
            nn.Conv2d(dims[0] // 2, dims[0], kernel_size=(2, 2), stride=(2, 2), padding=(0, 0), bias=False),
        )

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)

        prev_dim = dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = Conv2FormerStage(
                prev_dim,
                dims[i],
                kernel_size=kernel_size,
                mlp_ratio=mlp_ratios[i],
                drop_path=dpr[i],
                depth=depths[i],
                downsample=i > 0,
            )
            return_channels.append(dims[i])
            prev_dim = dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.Conv2d(dims[-1], 1280, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
            nn.GELU(),
            ChannelsFirstLayerNorm(1280, eps=1e-6),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = 1280
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = dims[0]
        self.encoding_size = 1280

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "conv2former_n",
    Conv2Former,
    config={"kernel_size": (7, 7), "dims": [64, 128, 256, 512], "depths": [2, 2, 8, 2], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "conv2former_t",
    Conv2Former,
    config={"kernel_size": (11, 11), "dims": [72, 144, 288, 576], "depths": [3, 3, 12, 3], "drop_path_rate": 0.15},
)
registry.register_model_config(
    "conv2former_s",
    Conv2Former,
    config={"kernel_size": (11, 11), "dims": [72, 144, 288, 576], "depths": [4, 4, 32, 4], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "conv2former_b",
    Conv2Former,
    config={"kernel_size": (11, 11), "dims": [96, 192, 384, 768], "depths": [4, 4, 34, 4], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "conv2former_l",
    Conv2Former,
    config={"kernel_size": (11, 11), "dims": [128, 256, 512, 1024], "depths": [4, 4, 48, 4], "drop_path_rate": 0.3},
)

registry.register_weights(
    "conv2former_n_il-common",
    {
        "description": "Conv2Former nano model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 54.3,
                "sha256": "47e93b34cfd7773f2a5c64bfe3c55a63e6c943dc1f4a130d5a4711ebdc96b632",
            }
        },
        "net": {"network": "conv2former_n", "tag": "il-common"},
    },
)



import os
from collections.abc import Callable
from typing import Any
from typing import Optional
from typing import cast

import numpy as np
import torch
from torchvision.datasets import DatasetFolder
from torchvision.datasets.folder import IMG_EXTENSIONS
from torchvision.datasets.folder import default_loader
from torchvision.datasets.folder import has_file_allowed_extension
from torchvision.io import ImageReadMode
from torchvision.io import decode_image
from torchvision.transforms.v2 import functional as F

from birder.common import fs_ops


def tv_loader(path: str) -> torch.Tensor:
    return decode_image(path, mode=ImageReadMode.RGB)


def default_is_valid_file(x: str) -> bool:
    return has_file_allowed_extension(x, IMG_EXTENSIONS)  # type: ignore[no-any-return]


def find_hierarchical_classes(
    directory: str, separator: str, is_valid_file: Optional[Callable[[str], bool]]
) -> tuple[list[str], dict[str, int]]:
    label_set = set()
    for root, _, files in os.walk(directory):
        rel_path = os.path.relpath(root, directory)
        if rel_path == ".":
            continue

        # Check if this directory contains any valid files
        if is_valid_file is not None:
            has_valid_files = any(is_valid_file(os.path.join(root, f)) for f in files)
        else:
            has_valid_files = True

        if has_valid_files is True:
            class_name = rel_path.replace(os.sep, separator)
            label_set.add(class_name)

    classes = sorted(list(label_set))
    class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}

    return (classes, class_to_idx)


class HierarchicalImageFolder(DatasetFolder):
    """
    A dataset class that supports hierarchical directory structures.

    Unlike the standard ImageFolder which uses only the immediate parent directory
    as the class label, this class creates labels from the full path hierarchy
    relative to the root directory.

    For example, given structure:
    root/
        dir1/
            subdir1/img1.jpeg -> label: "dir1_subdir1"
            subdir2/img2.jpeg -> label: "dir1_subdir2"
        dir2/
            subdir3/
                sub-subdir4/img3.jpeg -> label: "dir2_subdir3_sub-subdir4"
    """

    def __init__(
        self,
        root: str,
        loader: Callable[[str], Any] = default_loader,
        extensions: Optional[tuple[str, ...]] = None,
        transform: Optional[Callable[..., torch.Tensor]] = None,
        target_transform: Optional[Callable[..., torch.Tensor]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        allow_empty: bool = False,
        separator: str = "_",
    ):
        self.separator = separator
        if extensions is None:
            extensions = IMG_EXTENSIONS

        self.extensions = extensions
        super().__init__(
            root=root,
            loader=loader,
            extensions=extensions,
            transform=transform,
            target_transform=target_transform,
            is_valid_file=is_valid_file,
            allow_empty=allow_empty,
        )

    def _is_valid_file(
        self,
        file_path: str,
        extensions: Optional[tuple[str, ...]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ) -> bool:
        if is_valid_file is not None:
            return is_valid_file(file_path)

        if extensions is None:
            extensions = self.extensions

        return file_path.lower().endswith(extensions)

    def find_classes(self, directory: str) -> tuple[list[str], dict[str, int]]:
        return find_hierarchical_classes(directory, separator=self.separator, is_valid_file=self._is_valid_file)

    def make_dataset(  # pylint: disable=arguments-renamed
        self,
        directory: str,
        class_to_idx: dict[str, int],
        extensions: Optional[tuple[str, ...]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
        allow_empty: bool = False,
    ) -> list[tuple[str, int]]:
        directory = os.path.expanduser(directory)

        both_none = extensions is None and is_valid_file is None
        both_something = extensions is not None and is_valid_file is not None
        if both_none is True or both_something is True:
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")

        if extensions is not None:

            def is_valid_file(x: str) -> bool:  # pylint: disable=function-redefined
                return has_file_allowed_extension(x, extensions)  # type: ignore[no-any-return]

        is_valid_file = cast(Callable[[str], bool], is_valid_file)

        instances = []
        available_classes = set()
        for root, _, file_names in sorted(os.walk(directory, followlinks=True)):
            rel_path = os.path.relpath(root, directory)
            if rel_path == ".":
                continue

            class_name = rel_path.replace(os.sep, self.separator)
            if class_name not in class_to_idx:
                continue

            class_index = class_to_idx[class_name]

            # Add all valid files in this directory
            for fname in sorted(file_names):
                file_path = os.path.join(root, fname)
                if self._is_valid_file(file_path, extensions, is_valid_file):
                    instances.append((file_path, class_index))
                    if class_name not in available_classes:
                        available_classes.add(class_name)

        empty_classes = set(class_to_idx.keys()) - available_classes
        if empty_classes and not allow_empty:
            msg = f"Found no valid file for the classes {', '.join(sorted(empty_classes))}. "
            if extensions is not None:
                msg += f"Supported extensions are: {', '.join(extensions)}"

            raise FileNotFoundError(msg)

        return instances


class ImageListDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        samples: list[tuple[str, int]],
        transforms: Optional[Callable[..., torch.Tensor]] = None,
        loader: Callable[[str], torch.Tensor] = decode_image,
    ) -> None:
        super().__init__()
        self.transforms = transforms
        self.loader = loader

        # Avoid yielding Python objects
        # see: https://ppwwyyxx.com/blog/2022/Demystify-RAM-Usage-in-Multiprocess-DataLoader/
        (paths, labels) = list(zip(*samples))
        self.labels = np.array(labels, dtype=np.int32)
        self.paths = np.array(paths, dtype=np.bytes_)

    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any]:
        path = self.paths[index].decode("utf-8")
        label = self.labels[index].item()
        img = self.loader(path)
        if self.transforms is not None:
            sample = self.transforms(img)

        else:
            sample = img

        return (path, sample, label)

    def __len__(self) -> int:
        return len(self.paths)

    def __repr__(self) -> str:
        head = "Dataset " + self.__class__.__name__
        body = [f"Number of data points: {self.__len__()}"]
        if hasattr(self, "transforms") is True and self.transforms is not None:
            body += [repr(self.transforms)]

        lines = [head] + ["    " + line for line in body]

        return "\n".join(lines)


class ImageListDatasetWithSize(ImageListDataset):
    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any, list[int]]:  # type: ignore[override]
        path = self.paths[index].decode("utf-8")
        label = self.labels[index].item()
        img = self.loader(path)
        if self.transforms is not None:
            sample = self.transforms(img)

        else:
            sample = img

        return (path, sample, label, F.get_size(img))


def make_image_dataset(
    paths: list[str],
    class_to_idx: dict[str, int],
    transforms: Optional[Callable[..., torch.Tensor]] = None,
    loader: Callable[[str], torch.Tensor] = decode_image,
    *,
    return_orig_sizes: bool = False,
    hierarchical: bool = False,
) -> ImageListDataset:
    samples = fs_ops.collect_samples_from_paths(paths, class_to_idx=class_to_idx, hierarchical=hierarchical)
    if return_orig_sizes is True:
        dataset = ImageListDatasetWithSize(samples, transforms=transforms, loader=loader)
    else:
        dataset = ImageListDataset(samples, transforms=transforms, loader=loader)

    return dataset



import logging
import unittest

import torch

from birder.ops.msda import MultiScaleDeformableAttention
from birder.ops.msda import multi_scale_deformable_attention
from birder.ops.soft_nms import SoftNMS
from birder.ops.soft_nms import batched_soft_nms
from birder.ops.swattention import SWAttention_AV
from birder.ops.swattention import SWAttention_QK_RPB
from birder.ops.swattention import set_swattention_num_threads
from birder.ops.swattention import swattention_av
from birder.ops.swattention import swattention_qk_rpb

logging.disable(logging.CRITICAL)


class TestOps(unittest.TestCase):
    @unittest.skipUnless(torch.cuda.is_available(), "CUDA not available")
    def test_msda(self) -> None:
        device = torch.device("cuda")
        msda = MultiScaleDeformableAttention()
        self.assertTrue(msda.is_available)

        value = torch.rand(1, 34000, 8, 32, device=device)
        value_spatial_shapes = torch.tensor(
            [[160, 160], [80, 80], [40, 40], [20, 20]], dtype=torch.int64, device=device
        )
        value_level_start_index = torch.randint(10, (4,), dtype=torch.int64, device=device)
        sampling_locations = torch.rand(1, 34000, 8, 4, 4, 2, device=device)
        attention_weights = torch.rand(1, 34000, 8, 4, 4, device=device)
        im2col_step = 64

        op_kernel = msda(
            value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, im2col_step
        )
        fb_kernel = multi_scale_deformable_attention(
            value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, im2col_step
        )

        self.assertEqual(op_kernel.size(), fb_kernel.size())

    @unittest.skipUnless(torch.cuda.is_available(), "CUDA not available")  # CUDA not actually required
    def test_soft_nms(self) -> None:
        soft_nms = SoftNMS()
        self.assertTrue(soft_nms.is_available)

        boxes = torch.tensor(
            [
                [10, 10, 30, 30],
                [20, 20, 40, 40],  # Overlaps with first
                [50, 50, 70, 70],  # No overlap
                [15, 15, 35, 35],  # Overlaps with first two
            ],
            dtype=torch.float32,
        )
        scores = torch.tensor([0.9, 0.8, 0.7, 0.85])
        class_ids = torch.tensor([0, 0, 1, 0])

        (op_scores, op_keep) = soft_nms(boxes, scores.clone(), class_ids)
        (fb_scores, fb_keep) = batched_soft_nms(boxes, scores.clone(), class_ids)

        torch.testing.assert_close(op_scores, fb_scores, rtol=1e-5, atol=1e-6)
        torch.testing.assert_close(op_keep, fb_keep, rtol=0, atol=0)

        # Empty input
        empty_boxes = torch.empty(0, 4)
        empty_scores = torch.empty(0)
        empty_ids = torch.empty(0, dtype=torch.int64)

        (op_scores, op_keep) = soft_nms(empty_boxes, empty_scores, empty_ids)
        (fb_scores, fb_keep) = batched_soft_nms(empty_boxes, empty_scores, empty_ids)

        self.assertEqual(len(op_keep), 0)
        self.assertEqual(len(fb_keep), 0)

    @unittest.skipUnless(torch.cuda.is_available(), "CUDA not available")
    def test_swattention(self) -> None:
        device = torch.device("cuda")
        set_swattention_num_threads(32)

        # QK RPB
        swa_qk_rpb = SWAttention_QK_RPB()
        self.assertTrue(swa_qk_rpb.is_available)

        kv = torch.rand(1, 196, 192 * 2, device=device)
        q_norm_scaled = torch.rand(1, 2, 196, 96, device=device)
        relative_pos_bias_local = torch.rand(2, 9, device=device)
        padding_mask = torch.rand(196, 9, device=device) > 0.5
        num_heads = 2
        head_dim = 96
        H = 14
        W = 14
        window_size = 3
        local_len = 9

        (op_attn_local, op_v_local) = swa_qk_rpb(
            kv, q_norm_scaled, relative_pos_bias_local, padding_mask, num_heads, head_dim, window_size, local_len, H, W
        )
        (fb_attn_local, fb_v_local) = swattention_qk_rpb(
            kv, q_norm_scaled, relative_pos_bias_local, padding_mask, num_heads, head_dim, window_size, local_len, H, W
        )
        self.assertEqual(op_attn_local.size(), fb_attn_local.size())

        op_v_local = op_v_local.contiguous()
        fb_v_local = fb_v_local.contiguous()

        # AV
        swa_av = SWAttention_AV()
        self.assertTrue(swa_av.is_available)

        q_norm = torch.rand(1, 2, 196, 24, device=device)
        attn_local = torch.rand(1, 2, 196, 9, device=device)
        learnable_tokens = torch.rand(2, 24, 9, device=device)
        learnable_bias = torch.rand(2, 1, 9, device=device)

        op_x_local = swa_av(q_norm, attn_local, op_v_local, learnable_tokens, learnable_bias, window_size, H, W)
        fb_x_local = swattention_av(q_norm, attn_local, fb_v_local, learnable_tokens, learnable_bias)

        self.assertEqual(op_x_local.size(), fb_x_local.size())



"""
CSWin Transformer, adapted from
https://github.com/microsoft/CSWin-Transformer/blob/main/models/cswin.py

Paper "CSWin Transformer: A General Vision Transformer Backbone with Cross-Shaped Windows",
https://arxiv.org/abs/2107.00652

Changes from original:
* Split size based on image size (image size // 32 where applicable)
"""

# Reference license: MIT

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.vit import PatchEmbed


def img2windows(img: torch.Tensor, h_sp: int, w_sp: int) -> torch.Tensor:
    (B, C, H, W) = img.size()
    img_reshape = img.view(B, C, H // h_sp, h_sp, W // w_sp, w_sp)
    img_perm = img_reshape.permute(0, 2, 4, 3, 5, 1).contiguous().reshape(-1, h_sp * w_sp, C)

    return img_perm


def windows2img(img_splits_hw: torch.Tensor, h_sp: int, w_sp: int, H: int, W: int) -> torch.Tensor:
    B = int(img_splits_hw.shape[0] / (H * W / h_sp / w_sp))
    img = img_splits_hw.view(B, H // h_sp, W // w_sp, h_sp, w_sp, -1)
    img = img.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)

    return img


class LePEAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        resolution: tuple[int, int],
        idx: int,
        split_size: tuple[int, int],
        num_heads: int,
        attn_drop: float,
    ) -> None:
        super().__init__()
        self.resolution = resolution
        self.split_size = split_size
        self.idx = idx
        self.num_heads = num_heads
        head_dim = dim // num_heads

        self.scale = head_dim**-0.5
        self.assign_sp_shape()

        self.get_v = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)
        self.attn_drop = nn.Dropout(attn_drop)

    def assign_sp_shape(self) -> None:
        if self.idx == -1:
            self.h_sp = self.resolution[0]
            self.w_sp = self.resolution[1]
        elif self.idx == 0:
            self.h_sp = self.resolution[0]
            self.w_sp = self.split_size[1]
        elif self.idx == 1:
            self.h_sp = self.split_size[0]
            self.w_sp = self.resolution[1]
        else:
            raise ValueError("unsupported idx")

    def im2cswin(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, C) = x.size()
        x = x.transpose(-2, -1).contiguous().view(B, C, self.resolution[0], self.resolution[1])
        x = img2windows(x, self.h_sp, self.w_sp)
        x = x.reshape(-1, self.h_sp * self.w_sp, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3).contiguous()

        return x

    def get_lepe(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        (B, _, C) = x.size()
        H = self.resolution[0]
        W = self.resolution[1]
        x = x.transpose(-2, -1).contiguous().view(B, C, H, W)

        h_sp = self.h_sp
        w_sp = self.w_sp
        x = x.view(B, C, H // h_sp, h_sp, W // w_sp, w_sp)
        x = x.permute(0, 2, 4, 1, 3, 5).contiguous().reshape(-1, C, h_sp, w_sp)  # (B', C, H', W')

        lepe = self.get_v(x)  # (B', C, H', W')
        lepe = lepe.reshape(-1, self.num_heads, C // self.num_heads, h_sp * w_sp).permute(0, 1, 3, 2).contiguous()

        x = x.reshape(-1, self.num_heads, C // self.num_heads, h_sp * w_sp).permute(0, 1, 3, 2).contiguous()

        return (x, lepe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (q, k, v) = x.unbind(0)

        (B, _, C) = q.shape

        q = self.im2cswin(q)
        k = self.im2cswin(k)
        (v, lepe) = self.get_lepe(v)

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)  # B head N C @ B head C N --> B head N N
        attn = F.softmax(attn, dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v) + lepe
        x = x.transpose(1, 2).reshape(-1, self.h_sp * self.w_sp, C)  # B head N N @ B head N C

        x = windows2img(x, self.h_sp, self.w_sp, self.resolution[0], self.resolution[1]).view(B, -1, C)

        return x


class MergeBlock(nn.Module):
    def __init__(self, dim: int, dim_out: int, resolution: tuple[int, int]) -> None:
        super().__init__()
        self.conv = nn.Conv2d(dim, dim_out, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.norm = nn.LayerNorm(dim_out)
        self.resolution = resolution

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, C) = x.size()
        H = self.resolution[0]
        W = self.resolution[1]
        x = x.transpose(-2, -1).contiguous().view(B, C, H, W)
        x = self.conv(x)
        (B, C) = x.shape[:2]
        x = x.view(B, C, -1).transpose(-2, -1).contiguous()
        x = self.norm(x)

        return x


class CSWinBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        resolution: tuple[int, int],
        num_heads: int,
        split_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.norm1 = nn.LayerNorm(dim)

        if resolution[0] == split_size[0] and resolution[1] == split_size[1]:
            self.branch_num = 1
        else:
            self.branch_num = 2

        self.proj = nn.Linear(dim, dim)

        if self.branch_num == 1:
            self.attentions = nn.ModuleList(
                [
                    LePEAttention(
                        dim,
                        resolution=resolution,
                        idx=-1,
                        split_size=split_size,
                        num_heads=num_heads,
                        attn_drop=attn_drop,
                    )
                    for _ in range(self.branch_num)
                ]
            )
        else:
            self.attentions = nn.ModuleList(
                [
                    LePEAttention(
                        dim // 2,
                        resolution=resolution,
                        idx=i,
                        split_size=split_size,
                        num_heads=num_heads // 2,
                        attn_drop=attn_drop,
                    )
                    for i in range(self.branch_num)
                ]
            )

        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)
        self.norm2 = nn.LayerNorm(dim)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, C) = x.shape

        qkv = self.qkv(self.norm1(x)).reshape(B, -1, 3, C).permute(2, 0, 1, 3)
        if self.branch_num == 2:
            # Some workaround for TorchScript
            outputs = []
            qkv_list = (qkv[:, :, :, : C // 2], qkv[:, :, :, C // 2 :])
            for attn, qkv_sp in zip(self.attentions, qkv_list):
                outputs.append(attn(qkv_sp))

            attn_x = torch.concat(outputs, dim=2)
        else:
            attn_x = self.attentions[0](qkv)

        attn_x = self.proj(attn_x)
        x = x + self.drop_path(attn_x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class CSWinStage(nn.Module):
    def __init__(
        self,
        dim: int,
        resolution: tuple[int, int],
        num_heads: int,
        split_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: list[float],
        depth: int,
        downsample: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = MergeBlock(dim, dim * 2, resolution=(resolution[0] * 2, resolution[1] * 2))
            dim = dim * 2
        else:
            self.downsample = nn.Identity()

        blocks = []
        for i in range(depth):
            blocks.append(
                CSWinBlock(
                    dim,
                    resolution=resolution,
                    num_heads=num_heads,
                    split_size=split_size,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path[i],
                )
            )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


# pylint: disable=invalid-name
class CSWin_Transformer(DetectorBackbone):
    square_only = True

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        embed_dim: int = self.config["embed_dim"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        drop_path_rate: float = self.config["drop_path_rate"]
        mlp_ratio = 4.0
        self.split_size = [
            (1, 1),
            (2, 2),
            (int(self.size[0] / (2**5)), int(self.size[1] / (2**5))),
            (int(self.size[0] / (2**5)), int(self.size[1] / (2**5))),
        ]

        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, embed_dim, kernel_size=(7, 7), stride=(4, 4), padding=(2, 2)),
            PatchEmbed(),
            nn.LayerNorm(embed_dim),
        )

        num_stages = len(depths)
        curr_dim = embed_dim
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = CSWinStage(
                curr_dim,
                resolution=(self.size[0] // (2 ** (i + 2)), self.size[1] // (2 ** (i + 2))),
                num_heads=num_heads[i],
                split_size=self.split_size[i],
                mlp_ratio=mlp_ratio,
                qkv_bias=True,
                proj_drop=0.0,
                attn_drop=0.0,
                drop_path=dpr[i],
                depth=depths[i],
                downsample=i > 0,
            )
            if i > 0:
                curr_dim = curr_dim * 2

            return_channels.append(curr_dim)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.LayerNorm(curr_dim),
            Permute([0, 2, 1]),
            nn.AdaptiveAvgPool1d(output_size=1),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = curr_dim
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                (B, L, C) = x.size()
                H = int(math.sqrt(L))
                W = H
                out[name] = x.transpose(-2, -1).contiguous().view(B, C, H, W)

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        new_base = (new_size[0] // 4, new_size[1] // 4)
        idx = 0
        for stage in self.body.modules():
            if isinstance(stage, CSWinStage):
                for m in stage.modules():
                    if isinstance(m, MergeBlock):
                        m.resolution = (new_base[0] * 2, new_base[1] * 2)

                    elif isinstance(m, CSWinBlock):
                        for attn in m.attentions:
                            attn.resolution = new_base
                            attn.split_size = (new_size[0] // 32, new_size[1] // 32)
                            attn.assign_sp_shape()

                new_base = (new_base[0] // 2, new_base[1] // 2)
                self.split_size[idx] = (new_size[0] // 32, new_size[1] // 32)
                idx += 1


registry.register_model_config(
    "cswin_transformer_t",
    CSWin_Transformer,
    config={"embed_dim": 64, "depths": [1, 2, 21, 1], "num_heads": [2, 4, 8, 16], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "cswin_transformer_s",
    CSWin_Transformer,
    config={"embed_dim": 64, "depths": [2, 4, 32, 2], "num_heads": [2, 4, 8, 16], "drop_path_rate": 0.4},
)
registry.register_model_config(
    "cswin_transformer_b",
    CSWin_Transformer,
    config={"embed_dim": 96, "depths": [2, 4, 32, 2], "num_heads": [4, 8, 16, 32], "drop_path_rate": 0.5},
)
registry.register_model_config(
    "cswin_transformer_l",
    CSWin_Transformer,
    config={"embed_dim": 144, "depths": [2, 4, 32, 2], "num_heads": [6, 12, 24, 24], "drop_path_rate": 0.5},
)



"""
Swin Transformer, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/swin_transformer.py

Paper "Swin Transformer: Hierarchical Vision Transformer using Shifted Windows",
https://arxiv.org/abs/2103.14030

Changes from original:
* Window size based on image size (image size // 32)
"""

# Reference license: BSD 3-Clause

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.fx
import torch.nn.functional as F
from scipy import interpolate
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


def patch_merging_pad(x: torch.Tensor) -> torch.Tensor:
    (H, W, _) = x.shape[-3:]
    x = F.pad(x, (0, 0, 0, W % 2, 0, H % 2))
    x0 = x[..., 0::2, 0::2, :]  # ... H/2 W/2 C
    x1 = x[..., 1::2, 0::2, :]  # ... H/2 W/2 C
    x2 = x[..., 0::2, 1::2, :]  # ... H/2 W/2 C
    x3 = x[..., 1::2, 1::2, :]  # ... H/2 W/2 C
    x = torch.concat([x0, x1, x2, x3], dim=-1)  # ... H/2 W/2 4*C

    return x


torch.fx.wrap("patch_merging_pad")


def get_relative_position_bias(
    relative_position_bias_table: torch.Tensor,
    relative_position_index: torch.Tensor,
    window_size: tuple[int, int],
) -> torch.Tensor:
    N = window_size[0] * window_size[1]
    relative_position_bias = relative_position_bias_table[relative_position_index]
    relative_position_bias = relative_position_bias.view(N, N, -1)
    relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous().unsqueeze(0)

    return relative_position_bias


torch.fx.wrap("get_relative_position_bias")


# pylint: disable=too-many-locals
def shifted_window_attention(
    x: torch.Tensor,
    qkv_weight: torch.Tensor,
    proj_weight: torch.Tensor,
    relative_position_bias: torch.Tensor,
    window_size: tuple[int, int],
    num_heads: int,
    shift_size: tuple[int, int],
    qkv_bias: Optional[torch.Tensor] = None,
    proj_bias: Optional[torch.Tensor] = None,
    logit_scale: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    (B, H, W, C) = x.size()

    # Pad feature maps to multiples of window size
    pad_b = (window_size[0] - H % window_size[0]) % window_size[0]
    pad_r = (window_size[1] - W % window_size[1]) % window_size[1]
    x = F.pad(x, (0, 0, 0, pad_r, 0, pad_b))
    (_, pad_h, pad_w, _) = x.size()

    # If window size is larger than feature size, there is no need to shift window
    shift_size_w = shift_size[0]
    shift_size_h = shift_size[1]
    if window_size[0] >= pad_h:
        shift_size_h = 0
    if window_size[1] >= pad_w:
        shift_size_w = 0

    shift_size = (shift_size_w, shift_size_h)

    # Cyclic shift
    if sum(shift_size) > 0:
        x = torch.roll(x, shifts=(-shift_size[0], -shift_size[1]), dims=(1, 2))

    # Partition windows
    num_windows = (pad_h // window_size[0]) * (pad_w // window_size[1])
    x = x.view(B, pad_h // window_size[0], window_size[0], pad_w // window_size[1], window_size[1], C)
    x = x.permute(0, 1, 3, 2, 4, 5).reshape(B * num_windows, window_size[0] * window_size[1], C)  # B*nW, Ws*Ws, C

    # Multi-head attention
    if logit_scale is not None and qkv_bias is not None:
        qkv_bias = qkv_bias.clone()
        length = qkv_bias.numel() // 3
        qkv_bias[length : 2 * length].zero_()

    qkv = F.linear(x, qkv_weight, qkv_bias)  # pylint: disable=not-callable
    qkv = qkv.reshape(x.size(0), x.size(1), 3, num_heads, C // num_heads).permute(2, 0, 3, 1, 4)
    q = qkv[0]
    k = qkv[1]
    v = qkv[2]
    if logit_scale is not None:
        # Cosine attention
        attn = F.normalize(q, dim=-1) @ F.normalize(k, dim=-1).transpose(-2, -1)
        logit_scale = torch.clamp(logit_scale, max=math.log(100.0)).exp()
        attn = attn * logit_scale

    else:
        q = q * (C // num_heads) ** -0.5
        attn = q.matmul(k.transpose(-2, -1))

    # Add relative position bias
    attn = attn + relative_position_bias

    if sum(shift_size) > 0:
        # Generate attention mask
        attn_mask = x.new_zeros((pad_h, pad_w))
        h_slices = ((0, -window_size[0]), (-window_size[0], -shift_size[0]), (-shift_size[0], None))
        w_slices = ((0, -window_size[1]), (-window_size[1], -shift_size[1]), (-shift_size[1], None))
        count = 0
        for h in h_slices:
            for w in w_slices:
                attn_mask[h[0] : h[1], w[0] : w[1]] = count
                count += 1

        attn_mask = attn_mask.view(pad_h // window_size[0], window_size[0], pad_w // window_size[1], window_size[1])
        attn_mask = attn_mask.permute(0, 2, 1, 3).reshape(num_windows, window_size[0] * window_size[1])
        attn_mask = attn_mask.unsqueeze(1) - attn_mask.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))
        attn = attn.view(x.size(0) // num_windows, num_windows, num_heads, x.size(1), x.size(1))
        attn = attn + attn_mask.unsqueeze(1).unsqueeze(0)
        attn = attn.view(-1, num_heads, x.size(1), x.size(1))

    attn = F.softmax(attn, dim=-1)

    x = attn.matmul(v).transpose(1, 2).reshape(x.size(0), x.size(1), C)
    x = F.linear(x, proj_weight, proj_bias)  # pylint: disable=not-callable

    # Reverse windows
    x = x.view(B, pad_h // window_size[0], pad_w // window_size[1], window_size[0], window_size[1], C)
    x = x.permute(0, 1, 3, 2, 4, 5).reshape(B, pad_h, pad_w, C)

    # Reverse cyclic shift
    if sum(shift_size) > 0:
        x = torch.roll(x, shifts=(shift_size[0], shift_size[1]), dims=(1, 2))

    # Unpad features
    x = x[:, :H, :W, :].contiguous()

    return x


torch.fx.wrap("shifted_window_attention")


class PatchMerging(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = nn.LayerNorm(4 * dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = patch_merging_pad(x)
        x = self.norm(x)
        x = self.reduction(x)  # ... H/2 W/2 2*C

        return x


class ShiftedWindowAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        window_size: tuple[int, int],
        shift_size: tuple[int, int],
        num_heads: int,
        qkv_bias: bool = True,
        proj_bias: bool = True,
    ) -> None:
        super().__init__()
        if len(window_size) != 2 or len(shift_size) != 2:
            raise ValueError("window_size and shift_size must be of length 2")

        self.window_size = window_size
        self.shift_size = shift_size
        self.num_heads = num_heads

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)

        self.define_relative_position_bias_table()
        self.define_relative_position_index()

    def define_relative_position_bias_table(self) -> None:
        # Define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * self.window_size[0] - 1) * (2 * self.window_size[1] - 1), self.num_heads)
        )  # 2*Wh-1 * 2*Ww-1, nH
        nn.init.trunc_normal_(self.relative_position_bias_table, std=0.02)

    def define_relative_position_index(self) -> None:
        # Get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid(coords_h, coords_w, indexing="ij"))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.window_size[0] - 1  # Shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1).flatten()  # Wh*Ww*Wh*Ww
        self.relative_position_index = nn.Buffer(relative_position_index)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        relative_position_bias = get_relative_position_bias(
            self.relative_position_bias_table, self.relative_position_index, self.window_size
        )
        return shifted_window_attention(
            x,
            self.qkv.weight,
            self.proj.weight,
            relative_position_bias,
            self.window_size,
            self.num_heads,
            shift_size=self.shift_size,
            qkv_bias=self.qkv.bias,
            proj_bias=self.proj.bias,
        )


class SwinTransformerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: tuple[int, int],
        shift_size: tuple[int, int],
        mlp_ratio: float,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()

        self.norm1 = nn.LayerNorm(dim)
        self.attn = ShiftedWindowAttention(
            dim,
            window_size,
            shift_size,
            num_heads,
        )
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, inplace=None)

        for m in self.mlp.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.normal_(m.bias, std=1e-6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.stochastic_depth(self.attn(self.norm1(x)))
        x = x + self.stochastic_depth(self.mlp(self.norm2(x)))

        return x


# pylint: disable=invalid-name
class Swin_Transformer_v1(DetectorBackbone):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_ratio = 4.0
        patch_size = 4
        window_size = (int(self.size[0] / (2**5)), int(self.size[1] / (2**5)))
        embed_dim: int = self.config["embed_dim"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            nn.Conv2d(
                self.input_channels,
                embed_dim,
                kernel_size=(patch_size, patch_size),
                stride=(patch_size, patch_size),
                padding=(0, 0),
                bias=True,
            ),
            Permute([0, 2, 3, 1]),
            nn.LayerNorm(embed_dim, eps=1e-5),
        )

        total_stage_blocks = sum(depths)
        stage_block_id = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        layers = []
        for i_stage, depth in enumerate(depths):
            dim = embed_dim * 2**i_stage
            for i_layer in range(depth):
                # Adjust stochastic depth probability based on the depth of the stage block
                sd_prob = drop_path_rate * float(stage_block_id) / (total_stage_blocks - 1)
                if i_layer % 2 == 0:
                    shift_size = (0, 0)
                else:
                    shift_size = (window_size[0] // 2, window_size[1] // 2)

                layers.append(
                    SwinTransformerBlock(
                        dim,
                        num_heads[i_stage],
                        window_size=window_size,
                        shift_size=shift_size,
                        mlp_ratio=mlp_ratio,
                        stochastic_depth_prob=sd_prob,
                    )
                )
                stage_block_id += 1

            stages[f"stage{i_stage+1}"] = nn.Sequential(*layers)
            return_channels.append(dim)
            layers = []

            # Add patch merging layer
            if i_stage < (len(depths) - 1):
                layers.append(PatchMerging(dim))

        num_features = embed_dim * 2 ** (len(depths) - 1)
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.LayerNorm(num_features, eps=1e-5),
            Permute([0, 3, 1, 2]),  # B H W C -> B C H W
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x.permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        for m in self.body.modules():
            if isinstance(m, SwinTransformerBlock):
                old_window_size = (old_size[0] // (2**5), old_size[1] // (2**5))
                new_window_size = (new_size[0] // (2**5), new_size[1] // (2**5))
                m.attn.window_size = new_window_size
                shift_size_w = m.attn.shift_size[0]
                shift_size_h = m.attn.shift_size[1]
                if m.attn.shift_size[0] != 0:
                    shift_size_w = m.attn.window_size[0] // 2

                if m.attn.shift_size[1] != 0:
                    shift_size_h = m.attn.window_size[1] // 2

                m.attn.shift_size = (shift_size_w, shift_size_h)

                m.attn.define_relative_position_index()

                # Interpolate relative_position_bias_table, adapted from
                # https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/pos_embed_rel.py
                src_size = (2 * old_window_size[0] - 1, 2 * old_window_size[1] - 1)
                dst_size = (2 * new_window_size[0] - 1, 2 * new_window_size[1] - 1)
                rel_pos_bias = m.attn.relative_position_bias_table
                rel_pos_bias = rel_pos_bias.detach()
                num_attn_heads = rel_pos_bias.size(1)

                def _calc(src: int, dst: int) -> list[float]:
                    (left, right) = 1.01, 1.5
                    while right - left > 1e-6:
                        q = (left + right) / 2.0
                        gp = (1.0 - q ** (src // 2)) / (1.0 - q)  # Geometric progression
                        if gp > dst // 2:
                            right = q
                        else:
                            left = q

                    dis = []
                    cur = 1.0
                    for i in range(src // 2):
                        dis.append(cur)
                        cur += q ** (i + 1)

                    r_ids = [-_ for _ in reversed(dis)]
                    return r_ids + [0] + dis

                y = _calc(src_size[0], dst_size[0])
                x = _calc(src_size[1], dst_size[1])

                ty = dst_size[0] // 2.0
                tx = dst_size[1] // 2.0
                dy = torch.arange(-ty, ty + 0.1, 1.0)
                dx = torch.arange(-tx, tx + 0.1, 1.0)
                dxy = torch.meshgrid(dx, dy, indexing="ij")

                all_rel_pos_bias = []
                for i in range(num_attn_heads):
                    z = rel_pos_bias[:, i].view(src_size[0], src_size[1]).float()
                    rgi = interpolate.RegularGridInterpolator(
                        (x, y), z.numpy().T, method="cubic", bounds_error=False, fill_value=None
                    )
                    r = torch.Tensor(rgi(dxy)).T.contiguous().to(rel_pos_bias.device)

                    r = r.view(-1, 1)
                    all_rel_pos_bias.append(r)

                rel_pos_bias = torch.concat(all_rel_pos_bias, dim=-1)
                m.attn.relative_position_bias_table = nn.Parameter(rel_pos_bias)


registry.register_model_config(
    "swin_transformer_v1_t",
    Swin_Transformer_v1,
    config={
        "embed_dim": 96,
        "depths": [2, 2, 6, 2],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "swin_transformer_v1_s",
    Swin_Transformer_v1,
    config={
        "embed_dim": 96,
        "depths": [2, 2, 18, 2],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "swin_transformer_v1_b",
    Swin_Transformer_v1,
    config={
        "embed_dim": 128,
        "depths": [2, 2, 18, 2],
        "num_heads": [4, 8, 16, 32],
        "drop_path_rate": 0.5,
    },
)
registry.register_model_config(
    "swin_transformer_v1_l",
    Swin_Transformer_v1,
    config={
        "embed_dim": 192,
        "depths": [2, 2, 18, 2],
        "num_heads": [6, 12, 24, 48],
        "drop_path_rate": 0.5,
    },
)



import gzip
import logging
import os
from functools import cached_property
from typing import Any
from typing import Literal
from typing import Optional

import numpy as np
import numpy.typing as npt
import polars as pl
from rich.console import Console
from rich.table import Table
from rich.text import Text
from sklearn.metrics import accuracy_score
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
from sklearn.metrics import f1_score

from birder.conf import settings

logger = logging.getLogger(__name__)


def top_k_accuracy_score(y_true: npt.NDArray[Any], y_pred: npt.NDArray[np.float64], top_k: int) -> list[int]:
    """
    Returns all the sample indices which are in the top-k predictions
    """

    if len(y_true.shape) == 2:
        y_true = np.argmax(y_true, axis=1)

    (num_samples, _num_labels) = y_pred.shape
    indices: list[int] = []
    arg_sorted = np.argpartition(y_pred, -top_k, axis=1)[:, -top_k:]
    for i in range(num_samples):
        if y_true[i] in arg_sorted[i]:
            indices.append(i)

    # NumPy version, for some reason it's slightly slower...
    # y_true_expanded = y_true[:, np.newaxis]
    # matches = np.any(arg_sorted == y_true_expanded, axis=1)
    # indices = np.where(matches)[0].tolist()

    return indices


# pylint: disable=too-many-public-methods
class Results:
    """
    Classification result analysis class
    """

    num_desc_cols = 4

    def __init__(
        self,
        sample_list: list[str],
        labels: list[int],
        label_names: list[str],
        output: npt.NDArray[np.float32],
        predictions: Optional[npt.NDArray[np.int_]] = None,
        *,
        lazy: bool = True,
    ):
        """
        Initialize a result object

        Parameters
        ----------
        sample_list
            Sample names.
        labels
            The ground truth labels per sample.
        label_names
            Label names by order.
        output
            Probability of each class for each sample.
        predictions
            Prediction of each sample.
        lazy
            If True, metrics and computed properties will be lazily evaluated (computed only when accessed).
            If False, all metrics will be computed during initialization.
        """

        assert len(label_names) == len(output[0]), "Model output and label name list do not match"
        assert len(sample_list) == len(output), "Each output must have a sample name"

        if predictions is None:
            predictions = output.argmax(axis=1)

        names = [label_names[label] if label != settings.NO_LABEL else "" for label in labels]
        self._label_names = label_names
        self._lazy = lazy

        output_df = pl.DataFrame(
            {
                **{f"{i}": output[:, i].astype(np.float32) for i in range(output.shape[-1])},
            }
        )

        self._results_df = pl.DataFrame(
            {"sample": sample_list, "label": labels, "label_name": names, "prediction": predictions}
        )
        self._results_df = pl.concat([self._results_df, output_df], how="horizontal")
        self._results_df = self._results_df.sort("sample", descending=False)

        self._setup_metrics_and_flags()

    def _setup_metrics_and_flags(self) -> None:
        if np.all(self.labels == settings.NO_LABEL):
            self.missing_all_labels = True
        else:
            self.missing_all_labels = False

        # Calculate metrics
        if self.missing_all_labels is False:
            self.valid_idx = self.labels != settings.NO_LABEL
            self._valid_length: int = np.sum(self.valid_idx).item()
            accuracy: int = int(
                accuracy_score(self.labels[self.valid_idx], self.predictions[self.valid_idx], normalize=False)
            )
            self._num_mistakes = self._valid_length - accuracy
            self._accuracy = accuracy / self._valid_length

            if self._lazy is False:
                _ = self.top_k
                _ = self.confusion_matrix

        else:
            self.valid_idx = np.array([], dtype=bool)  # Empty boolean array
            self._valid_length = 0
            self._num_mistakes = np.nan
            self._accuracy = np.nan

    def __len__(self) -> int:
        return len(self._results_df)

    def __repr__(self) -> str:
        head = self.__class__.__name__
        body = [
            f"Number of samples: {len(self)}",
            f"Number of valid samples: {self._valid_length}",
        ]

        if self.missing_all_labels is False:
            body.append(f"Accuracy: {self.accuracy:.4f}")

        lines = [head] + ["    " + line for line in body]

        return "\n".join(lines)

    @cached_property
    def _top_k_indices(self) -> list[int]:
        return top_k_accuracy_score(self.labels, self.output, top_k=settings.TOP_K)

    @property
    def labels(self) -> npt.NDArray[np.int_]:
        return self._results_df["label"].to_numpy()

    @property
    def label_names(self) -> list[str]:
        return self._label_names

    @cached_property
    def unique_labels(self) -> npt.NDArray[np.int_]:
        return np.unique(np.concatenate([self.labels[self.valid_idx], self.predictions[self.valid_idx]], axis=0))

    @property
    def missing_labels(self) -> bool:
        if settings.NO_LABEL in self.labels:
            return True

        return False

    @property
    def output(self) -> npt.NDArray[np.float64]:
        return self.output_df.to_numpy()

    @property
    def output_df(self) -> pl.DataFrame:
        return self._results_df.select(pl.all().exclude(self._results_df.columns[: Results.num_desc_cols]))

    @property
    def predictions(self) -> npt.NDArray[np.int_]:
        return self._results_df["prediction"].to_numpy()

    @property
    def prediction_names(self) -> pl.Series:
        prediction_names = pl.Series("prediction_names", self._label_names)
        return prediction_names[self.predictions]

    @property
    def mistakes(self) -> pl.DataFrame:
        return self._results_df.filter(pl.col("label") != pl.col("prediction"))

    @property
    def out_of_top_k(self) -> pl.DataFrame:
        return self._results_df.with_row_index().filter(~pl.col("index").is_in(self._top_k_indices)).drop("index")

    @property
    def num_out_of_top_k(self) -> int:
        return self._valid_length - len(self._top_k_indices)

    @property
    def accuracy(self) -> float:
        return self._accuracy  # type: ignore[no-any-return]

    @cached_property
    def top_k(self) -> float:
        return len(self._top_k_indices) / self._valid_length

    @property
    def macro_f1_score(self) -> float:
        return f1_score(self.labels[self.valid_idx], self.predictions[self.valid_idx], average="macro")  # type: ignore

    @cached_property
    def confusion_matrix(self) -> npt.NDArray[np.int_]:
        return confusion_matrix(self.labels[self.valid_idx], self.predictions[self.valid_idx])  # type: ignore

    def most_confused_pairs(self, n: int = 10) -> pl.DataFrame:
        cnf_matrix = self.confusion_matrix.copy()
        np.fill_diagonal(cnf_matrix, -1)
        top_indices = np.argsort(cnf_matrix.ravel())[-n:][::-1]
        class_names = [self.label_names[label_idx] for label_idx in self.unique_labels]

        data = []
        for flat_idx in top_indices:
            idx = np.unravel_index(flat_idx, cnf_matrix.shape)
            if cnf_matrix[idx] == 0:
                break

            data.append(
                {
                    "predicted": class_names[idx[1]],
                    "actual": class_names[idx[0]],
                    "amount": cnf_matrix[idx],
                    "reverse": cnf_matrix[idx[::-1]],
                }
            )

        return pl.DataFrame(data)

    def to_dataframe(self) -> pl.DataFrame:
        return self._results_df.clone()

    def detailed_report(self) -> pl.DataFrame:
        """
        Returns a detailed classification report with per-class metrics
        """

        raw_report_dict: dict[str, dict[str, float]] = classification_report(
            self.labels[self.valid_idx], self.predictions[self.valid_idx], output_dict=True, zero_division=0
        )
        del raw_report_dict["accuracy"]
        del raw_report_dict["macro avg"]
        del raw_report_dict["weighted avg"]

        # Pre-compute row and column sums for the confusion matrix
        cm_row_sums = np.sum(self.confusion_matrix, axis=1)
        cm_col_sums = np.sum(self.confusion_matrix, axis=0)

        row_list = []
        for class_idx, metrics in raw_report_dict.items():
            class_num = int(class_idx)

            # Skip metrics on classes we did not predict
            if metrics["support"] == 0:
                continue

            # Cast to int
            metrics["support"] = int(metrics["support"])

            # Get label name
            label_name = self._label_names[class_num]

            # Calculate additional metrics
            item_index = np.asarray(self.unique_labels == class_num).nonzero()[0][0]
            false_negative = cm_row_sums[item_index] - self.confusion_matrix[item_index][item_index]
            false_positive = cm_col_sums[item_index] - self.confusion_matrix[item_index][item_index]

            # Save metrics
            row_list.append(
                {
                    "Class": class_num,
                    "Class name": label_name,
                    "Precision": metrics["precision"],
                    "Recall": metrics["recall"],
                    "F1-score": metrics["f1-score"],
                    "Samples": metrics["support"],
                    "False negative": false_negative,
                    "False positive": false_positive,
                }
            )

        return pl.DataFrame(row_list)

    def log_short_report(self) -> None:
        """
        Log using the Python logging module a short metrics summary
        """

        report_df = self.detailed_report()
        lowest_precision = report_df[report_df["Precision"].arg_min()]  # type: ignore[index]
        lowest_recall = report_df[report_df["Recall"].arg_min()]  # type: ignore[index]
        highest_precision = report_df[report_df["Precision"].arg_max()]  # type: ignore[index]
        highest_recall = report_df[report_df["Recall"].arg_max()]  # type: ignore[index]

        logger.info(f"Accuracy {self.accuracy:.4f} on {self._valid_length} samples ({self._num_mistakes} mistakes)")
        logger.info(
            f"Top-{settings.TOP_K} accuracy {self.top_k:.4f} on {self._valid_length} samples "
            f"({self.num_out_of_top_k} samples out of top-{settings.TOP_K})"
        )

        logger.info(
            f"Lowest precision {lowest_precision['Precision'][0]:.4f} for '{lowest_precision['Class name'][0]}' "
            f"({lowest_precision['False negative'][0]} false negatives, "
            f"{lowest_precision['False positive'][0]} false positives)"
        )
        logger.info(
            f"Lowest recall {lowest_recall['Recall'][0]:.4f} for '{lowest_recall['Class name'][0]}' "
            f"({lowest_recall['False negative'][0]} false negatives, "
            f"{lowest_recall['False positive'][0]} false positives)"
        )

        logger.info(
            f"Highest precision {highest_precision['Precision'][0]:.4f} for '{highest_precision['Class name'][0]}' "
            f"({highest_precision['False negative'][0]} false negatives, "
            f"{highest_precision['False positive'][0]} false positives)"
        )
        logger.info(
            f"Highest recall {highest_recall['Recall'][0]:.4f} for '{highest_recall['Class name'][0]}' "
            f"({highest_recall['False negative'][0]} false negatives, "
            f"{highest_recall['False positive'][0]} false positives)"
        )
        if self.missing_labels is True:
            logger.warning(
                f"{len(self) - self._valid_length} of the samples did not have labels, metrics calculated only on "
                f"{self._valid_length} out of total {len(self)} samples"
            )

    def pretty_print(
        self,
        sort_by: Literal["class", "precision", "recall", "f1-score"] = "class",
        order: Literal["ascending", "descending"] = "ascending",
        n: Optional[int] = None,
    ) -> None:
        console = Console()

        table = Table(show_header=True, header_style="bold dark_magenta")
        table.add_column("Class")
        table.add_column("Class name", style="dim")
        table.add_column("Precision", justify="right")
        table.add_column("Recall", justify="right")
        table.add_column("F1-score", justify="right")
        table.add_column("Samples", justify="right")
        table.add_column("False negative", justify="right")
        table.add_column("False positive", justify="right")

        report_df = self.detailed_report()
        report_df = report_df.sort(sort_by.capitalize(), descending=order == "descending")
        if n is not None:
            report_df = report_df[:n]

        fn_cutoff = report_df["False negative"].quantile(0.95)
        fp_cutoff = report_df["False positive"].quantile(0.95)

        for row in report_df.iter_rows(named=True):
            recall_msg = f"{row['Recall']:.4f}"
            if row["Recall"] < 0.75:
                recall_msg = "[red1]" + recall_msg + "[/red1]"

            elif row["Recall"] < 0.9:
                recall_msg = "[dark_orange]" + recall_msg + "[/dark_orange]"

            f1_msg = f"{row['F1-score']:.4f}"
            if row["F1-score"] == 1.0:
                f1_msg = "[green]" + f1_msg + "[/green]"

            fn_msg = f"{row['False negative']}"
            if row["False negative"] > fn_cutoff:
                fn_msg = "[underline]" + fn_msg + "[/underline]"

            fp_msg = f"{row['False positive']}"
            if row["False positive"] > fp_cutoff:
                fp_msg = "[underline]" + fp_msg + "[/underline]"

            table.add_row(
                f"{row['Class']}",
                row["Class name"],
                f"{row['Precision']:.4f}",
                recall_msg,
                f1_msg,
                f"{row['Samples']}",
                fn_msg,
                fp_msg,
            )

        console.print(table)

        accuracy_text = Text()
        accuracy_text.append(f"Accuracy {self.accuracy:.4f} on {self._valid_length} samples (")
        accuracy_text.append(f"{self._num_mistakes}", style="bold")
        accuracy_text.append(" mistakes)")

        top_k_text = Text()
        top_k_text.append(f"Top-{settings.TOP_K} accuracy {self.top_k:.4f} on {self._valid_length} samples (")
        top_k_text.append(f"{self.num_out_of_top_k}", style="bold")
        top_k_text.append(f" samples out of top-{settings.TOP_K})")

        console.print(accuracy_text)
        console.print(top_k_text)
        if self.missing_labels is True:
            console.print(
                "[bold][bright_red]NOTICE[/bright_red][/bold]: "
                f"{len(self) - self._valid_length} of the samples did not have labels, metrics calculated only on "
                f"{self._valid_length} out of total {len(self)} samples"
            )

    def filter_by_labels(self, target_label_indices: list[int]) -> "Results":
        filtered_df = self._results_df.filter(pl.col("label").is_in(target_label_indices))

        samples = filtered_df["sample"].to_list()
        labels = filtered_df["label"].to_list()
        predictions = filtered_df["prediction"].to_numpy()
        output = filtered_df[:, Results.num_desc_cols :].to_numpy()

        return Results(samples, labels, self._label_names, output, predictions, lazy=self._lazy)

    def save(self, name: str, append: bool = False) -> None:
        """
        Save results object to file

        Parameters
        ----------
        name
            Output file name.
        append
            Append result data to existing results file.
        """

        if settings.RESULTS_DIR.exists() is False:
            logger.info(f"Creating {settings.RESULTS_DIR} directory...")
            settings.RESULTS_DIR.mkdir(parents=True)

        results_path = settings.RESULTS_DIR.joinpath(name)
        if append is False:
            logger.info(f"Saving results at {results_path}")

            # Write label names list
            with open(results_path, "w", encoding="utf-8") as handle:
                handle.write("," * Results.num_desc_cols)
                handle.write(",".join(self._label_names))
                handle.write(os.linesep)

            # Write the data frame
            with open(results_path, "a", encoding="utf-8") as handle:
                self._results_df.write_csv(handle)
        else:
            logger.info(f"Adding results to {results_path}")
            with open(results_path, "a", encoding="utf-8") as handle:
                self._results_df.write_csv(handle, include_header=False)

    @staticmethod
    def load(path: str, lazy: bool = True) -> "Results":
        """
        Load results object from file

        Parameters
        ----------
        path
            Path to load from.
        lazy
            If True, metrics and computed properties will be lazily evaluated (computed only when accessed).
            If False, all metrics will be computed during initialization.
        """

        # Read label names
        if path.endswith(".gz") is True:
            with gzip.open(path, "rt", encoding="utf-8") as handle:
                label_names = handle.readline().rstrip(os.linesep).split(",")
                label_names = label_names[Results.num_desc_cols :]
        else:
            with open(path, "r", encoding="utf-8") as handle:
                label_names = handle.readline().rstrip(os.linesep).split(",")
                label_names = label_names[Results.num_desc_cols :]

        # Read the data frame
        schema_overrides = {
            "sample": pl.String,
            "label": pl.Int32,
            "label_name": pl.String,
            "prediction": pl.Int32,
            **{str(i): pl.Float32 for i in range(len(label_names))},
        }
        results_df = pl.read_csv(path, skip_rows=1, schema_overrides=schema_overrides)
        return Results(
            results_df["sample"].to_list(),
            results_df["label"].to_list(),
            label_names,
            results_df[:, Results.num_desc_cols :].to_numpy(),
            results_df["prediction"].to_numpy(),
            lazy=lazy,
        )


class SparseResults(Results):
    """
    Memory-efficient classification result analysis class that stores only top-k probabilities.
    """

    def __init__(  # pylint: disable=super-init-not-called
        self,
        sample_list: list[str],
        labels: list[int],
        label_names: list[str],
        output: Optional[npt.NDArray[np.float32]],
        predictions: Optional[npt.NDArray[np.int_]] = None,
        *,
        lazy: bool = True,
        sparse_k: int = 10,
        sparse_probs: Optional[npt.NDArray[np.float32]] = None,
        sparse_indices: Optional[npt.NDArray[np.int_]] = None,
    ):
        """
        Initialize a sparse result object

        Parameters
        ----------
        sample_list
            Sample names.
        labels
            The ground truth labels per sample.
        label_names
            Label names by order.
        output
            Probability of each class for each sample.
        predictions
            Prediction of each sample.
        lazy
            If True, metrics and computed properties will be lazily evaluated (computed only when accessed).
            If False, all metrics will be computed during initialization.
        sparse_k
            Number of top probabilities to keep per sample.
        sparse_probs
            Pre-extracted top-k probabilities (used when loading from file).
        sparse_indices
            Pre-extracted top-k indices (used when loading from file).
        """

        if output is not None:
            assert len(label_names) == len(output[0]), "Model output and label name list do not match"
            assert len(sample_list) == len(output), "Each output must have a sample name"
            assert sparse_k < output.shape[1], "sparse_k must be smaller than the number of classes"
            assert sparse_k >= settings.TOP_K, "sparse_k must be larger than top-k being calculated"

            self._sparse_k = min(sparse_k, len(label_names))
            if predictions is None:
                predictions = output.argmax(axis=1)

            # Extract and store only top-k probabilities and their indices
            self._extract_sparse_probabilities(output)

        elif sparse_indices is not None and sparse_probs is not None:
            assert len(sample_list) == len(labels), "Each label must have a sample name"

            self._sparse_indices = sparse_indices
            self._sparse_probs = sparse_probs
            self._sparse_k = sparse_indices.shape[1]  # Infer sparse_k from loaded data
            if predictions is None:
                predictions = sparse_indices[:, 0]

        else:
            raise ValueError("Either 'output' or 'sparse_indices' and 'sparse_probs' must be provided")

        names = [label_names[label] if label != settings.NO_LABEL else "" for label in labels]
        self._label_names = label_names
        self._lazy = lazy

        self._results_df = pl.DataFrame(
            {"sample": sample_list, "label": labels, "label_name": names, "prediction": predictions}
        )
        self._results_df = self._results_df.sort("sample", descending=False)

        self._setup_metrics_and_flags()

    def _extract_sparse_probabilities(self, output: npt.NDArray[np.float32]) -> None:
        top_k_indices = np.argpartition(output, -self._sparse_k, axis=1)[:, -self._sparse_k :]
        top_k_probs = np.take_along_axis(output, top_k_indices, axis=1)

        # Sort within each row for easier access (highest prob first)
        sort_idx = np.argsort(-top_k_probs, axis=1)
        self._sparse_indices = np.take_along_axis(top_k_indices, sort_idx, axis=1).astype(np.int32)
        self._sparse_probs = np.take_along_axis(top_k_probs, sort_idx, axis=1)

    def filter_by_labels(self, target_label_indices: list[int]) -> "SparseResults":
        indexed_df = self._results_df.with_row_index("original_index")
        filtered_indexed_df = indexed_df.filter(pl.col("label").is_in(target_label_indices))

        original_row_indices = filtered_indexed_df["original_index"].to_numpy()

        samples = filtered_indexed_df["sample"].to_list()
        labels = filtered_indexed_df["label"].to_list()
        predictions = filtered_indexed_df["prediction"].to_numpy()

        sparse_probs = self._sparse_probs[original_row_indices]
        sparse_indices = self._sparse_indices[original_row_indices]

        return SparseResults(
            samples,
            labels,
            self._label_names,
            output=None,
            predictions=predictions,
            lazy=self._lazy,
            sparse_k=self._sparse_k,
            sparse_probs=sparse_probs,
            sparse_indices=sparse_indices,
        )

    def save(self, name: str, append: bool = False) -> None:
        """
        Save sparse results object to file
        """

        if settings.RESULTS_DIR.exists() is False:
            logger.info(f"Creating {settings.RESULTS_DIR} directory...")
            settings.RESULTS_DIR.mkdir(parents=True)

        results_path = settings.RESULTS_DIR.joinpath(name)

        # Prepare sparse data for DataFrame
        sparse_dict = {}
        for k_val in range(self._sparse_k):
            sparse_dict[f"prob_{k_val}"] = self._sparse_probs[:, k_val].astype(np.float32)
            sparse_dict[f"idx_{k_val}"] = self._sparse_indices[:, k_val].astype(np.int32)

        sparse_df = pl.DataFrame(sparse_dict)
        sparse_results_df = pl.concat([self._results_df, sparse_df], how="horizontal")

        if append is False:
            logger.info(f"Saving results at {results_path}")

            # Write label names list
            with open(results_path, "w", encoding="utf-8") as handle:
                handle.write("," * Results.num_desc_cols)
                handle.write(",".join(self._label_names))
                handle.write(os.linesep)

            # Write the data frame
            with open(results_path, "a", encoding="utf-8") as handle:
                sparse_results_df.write_csv(handle)
        else:
            logger.info(f"Adding results to {results_path}")
            with open(results_path, "a", encoding="utf-8") as handle:
                sparse_results_df.write_csv(handle, include_header=False)

    @property
    def output(self) -> npt.NDArray[np.float64]:
        raise NotImplementedError("Output not defined in a SparseResults object")

    @property
    def output_df(self) -> pl.DataFrame:
        raise NotImplementedError("Output not defined in a SparseResults object")

    @cached_property
    def _top_k_indices(self) -> list[int]:
        indices: list[int] = []
        for i in range(len(self)):
            if self.labels[i] in self._sparse_indices[i][: settings.TOP_K]:
                indices.append(i)

        return indices

    @staticmethod
    def load(path: str, lazy: bool = True, sparse_k: int = 10) -> "SparseResults":
        """
        Load results object from file

        This method automatically detects whether the file is in sparse format
        (only top-k probabilities and corresponding class indices stored) or regular format
        (full probability distribution stored).

        Parameters
        ----------
        path
            Path to the saved results file (CSV or gzipped CSV).
        lazy
            If True, metrics and computed properties will be lazily evaluated (computed only when accessed).
            If False, all metrics will be computed during initialization.
        sparse_k
            Number of top probabilities to keep per sample when loading from a regular (non-sparse) file.
            For sparse files, this value is ignored.
        """

        (label_names, detected_sparse_k) = detect_file_format(path)

        if detected_sparse_k is not None:
            schema_overrides = {
                "sample": pl.String,
                "label": pl.Int32,
                "label_name": pl.String,
                "prediction": pl.Int32,
                **{f"prob_{i}": pl.Float32 for i in range(detected_sparse_k)},
                **{f"idx_{i}": pl.Int32 for i in range(detected_sparse_k)},
            }
            results_df = pl.read_csv(path, skip_rows=1, schema_overrides=schema_overrides)

            probs = np.stack([results_df[f"prob_{i}"].to_numpy() for i in range(detected_sparse_k)], axis=1)
            indices = np.stack([results_df[f"idx_{i}"].to_numpy() for i in range(detected_sparse_k)], axis=1)

            return SparseResults(
                results_df["sample"].to_list(),
                results_df["label"].to_list(),
                label_names,
                output=None,
                predictions=results_df["prediction"].to_numpy(),
                lazy=lazy,
                sparse_k=detected_sparse_k,
                sparse_probs=probs,
                sparse_indices=indices,
            )

        # Read the data frame
        schema_overrides = {
            "sample": pl.String,
            "label": pl.Int32,
            "label_name": pl.String,
            "prediction": pl.Int32,
            **{str(i): pl.Float32 for i in range(len(label_names))},
        }
        results_df = pl.read_csv(path, skip_rows=1, schema_overrides=schema_overrides)
        return SparseResults(
            results_df["sample"].to_list(),
            results_df["label"].to_list(),
            label_names,
            results_df[:, Results.num_desc_cols :].to_numpy(),
            results_df["prediction"].to_numpy(),
            lazy=lazy,
            sparse_k=sparse_k,
        )


def detect_file_format(path: str) -> tuple[list[str], Optional[int]]:
    """
    Detect file format and extract metadata from results file

    Parameters
    ----------
    path
        Path to the results file (CSV or gzipped CSV)

    Returns
    -------
    A tuple containing:
    - label_names: List of label names extracted from first line
    - sparse_k: Number of top-k probabilities if sparse format, None if regular format
    """

    # Read label names and peek at header to detect format
    if path.endswith(".gz") is True:
        with gzip.open(path, "rt", encoding="utf-8") as handle:
            label_names = handle.readline().rstrip(os.linesep).split(",")
            label_names = label_names[Results.num_desc_cols :]

            # Peek at the header line to detect format
            header_cols = handle.readline().rstrip(os.linesep).split(",")
    else:
        with open(path, "r", encoding="utf-8") as handle:
            label_names = handle.readline().rstrip(os.linesep).split(",")
            label_names = label_names[Results.num_desc_cols :]

            # Peek at the header line to detect format
            header_cols = handle.readline().rstrip(os.linesep).split(",")

    # Detect format by looking for sparse-specific columns
    has_prob_col = any(col.startswith("prob_") for col in header_cols)
    has_idx_col = any(col.startswith("idx_") for col in header_cols)
    is_sparse_format = has_prob_col and has_idx_col

    if is_sparse_format is True:
        # Count the number of prob_ columns to determine sparse_k
        sparse_k = len([col for col in header_cols if col.startswith("prob_")])
        return (label_names, sparse_k)

    return (label_names, None)


def load_results(path: str, lazy: bool = True) -> Results | SparseResults:
    """
    Automatically detect and load results from either regular or sparse format files

    This function examines the file structure to determine whether it contains
    regular results (full probability distributions) or sparse results (only top-k
    probabilities and indices), then loads using the appropriate class.

    Parameters
    ----------
    path
        Path to the results file (CSV or gzipped CSV)
    lazy
        If True, metrics and computed properties will be lazily evaluated.
        If False, all metrics will be computed during initialization.

    Returns
    -------
    Loaded results object of the appropriate type, Results or SparseResults

    Examples
    --------
    >>> results = load_results("model_results.csv")
    >>> type(results)
    <class 'birder.results.classification.Results'>

    >>> sparse_results = load_results("model_results_sparse.csv")
    >>> type(sparse_results)
    <class 'birder.results.classification.SparseResults'>
    """

    (_, sparse_k) = detect_file_format(path)

    # Load using appropriate class
    if sparse_k is not None:
        return SparseResults.load(path, lazy=lazy)

    return Results.load(path, lazy=lazy)


def compare_results(results_dict: dict[str, Results | SparseResults], include_top_k: bool = True) -> pl.DataFrame:
    result_list = []
    for name, results in results_dict.items():
        result_entry = {
            "File name": name,
            "Accuracy": results.accuracy,
        }
        if include_top_k is True:
            result_entry[f"Top-{settings.TOP_K} accuracy"] = results.top_k

        result_entry.update(
            {
                "Macro F1-score": results.macro_f1_score,
                "Samples": len(results),
                "Mistakes": results._num_mistakes,  # pylint: disable=protected-access
            }
        )
        result_list.append(result_entry)

    return pl.DataFrame(result_list)



import argparse
import json
import logging
import os
from datetime import date
from pathlib import Path
from typing import Any

from tqdm import tqdm

import birder
from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.conf import settings

try:
    from defusedxml.ElementTree import parse

    _HAS_DEFUSEDXML = True
except ImportError:
    _HAS_DEFUSEDXML = False

logger = logging.getLogger(__name__)


def _create_annotation(
    points: tuple[float, float, float, float], label: int, image_id: int, annotation_id: int
) -> dict[str, Any]:
    annotation: dict[str, Any] = {}
    annotation["iscrowd"] = 0
    annotation["image_id"] = image_id

    # Bounding box in (x, y, w, h) format
    (xmin, ymin, xmax, ymax) = points
    assert xmax > xmin and ymax > ymin
    w = xmax - xmin
    h = ymax - ymin
    annotation["bbox"] = [xmin, ymin, w, h]
    annotation["category_id"] = label
    annotation["id"] = annotation_id

    return annotation


# pylint: disable=too-many-locals
def voc_to_coco(args: argparse.Namespace) -> None:
    prefix = str(settings.DETECTION_DATA_PATH) + "/"
    target_path = Path(args.ann_dir).parent.joinpath(f"{Path(args.class_file).stem}_coco.json")
    if target_path.exists() is True:
        logger.warning(f"{target_path} already exists, aborting...")

    class_to_idx = fs_ops.read_class_file(args.class_file)
    class_to_idx = lib.detection_class_to_idx(class_to_idx)

    img_path = Path(os.path.relpath(args.data_path, start=Path(args.ann_dir).parent))

    image_list = []
    annotation_list = []
    annotation_id = 0
    for idx, xml_path in tqdm(enumerate(fs_ops.file_iter(args.ann_dir, extensions=[".xml"])), leave=False):
        ann_tree = parse(xml_path)
        ann_root = ann_tree.getroot()

        size = ann_root.find("size")
        width = int(size.findtext("width"))
        height = int(size.findtext("height"))

        filename = ann_root.findtext("filename")
        image_path = img_path.joinpath(filename)

        skip = False
        for obj in ann_root.findall("object"):
            label = obj.findtext("name")
            if label not in class_to_idx:
                logger.debug(f"Found unknown label: {label}, skipping image {img_path}")
                skip = True
                break

        if skip is True:
            continue

        image = {
            "id": idx,
            "width": width,
            "height": height,
            "file_name": os.path.normpath(image_path).removeprefix(prefix),
        }
        image_list.append(image)

        for obj in ann_root.findall("object"):
            label = obj.findtext("name")
            if label not in class_to_idx:
                logger.error(f"Found unknown label: {label}, aborting...")
                raise SystemExit(1)

            bbox = obj.find("bndbox")

            # VOC starts from (1, 1) instead of (0, 0)
            xmin = int(float(bbox.findtext("xmin"))) - 1
            ymin = int(float(bbox.findtext("ymin"))) - 1
            xmax = int(float(bbox.findtext("xmax")))
            ymax = int(float(bbox.findtext("ymax")))
            annotation_list.append(
                _create_annotation((xmin, ymin, xmax, ymax), class_to_idx[label], idx, annotation_id)
            )
            annotation_id += 1

    # Create categories
    category_list = []
    for class_name, class_id in class_to_idx.items():
        category_list.append(
            {
                "supercategory": class_name,
                "id": class_id,
                "name": class_name,
            }
        )

    # Create COCO format dictionary
    coco: dict[str, Any] = {}
    coco["info"] = {
        "version": "1.0",
        "birder_version": birder.__version__,
        "year": date.today().year,
        "date_created": date.today().isoformat(),
    }
    coco["images"] = image_list
    coco["categories"] = category_list
    coco["annotations"] = annotation_list

    # Save
    logger.info(f"Saving COCO file at {target_path}...")
    with open(target_path, "w", encoding="utf-8") as handle:
        json.dump(coco, handle, indent=2)

    logger.info(f"Written {len(image_list)} images with {len(annotation_list)} annotations")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "voc-to-coco",
        allow_abbrev=False,
        help="convert voc detection annotations to coco format",
        description="convert voc detection annotations to coco format",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools voc-to-coco --class-file public_datasets_metadata/voc-classes.txt "
            "--ann-dir ~/Datasets/VOC2012/Annotations ~/Datasets/VOC2012/JPEGImages\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--class-file", type=str, required=True, help="class list file")
    subparser.add_argument("--ann-dir", type=str, required=True, help="annotations directory")
    subparser.add_argument("data_path", help="image directory path")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    assert _HAS_DEFUSEDXML, "'pip install defusedxml' to use voc-to-coco"
    voc_to_coco(args)



from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
import torch.amp
from PIL import Image
from torch.utils.data import DataLoader
from tqdm import tqdm

from birder.conf import settings
from birder.data.transforms.detection import InferenceTransform


def infer_image(
    net: torch.nn.Module | torch.ScriptModule,
    sample: Image.Image | str,
    transform: Callable[..., torch.Tensor],
    device: Optional[torch.device] = None,
    score_threshold: Optional[float] = None,
    **kwargs: Any,
) -> dict[str, torch.Tensor]:
    """
    Perform inference on a single image

    This convenience function allows for quick, one-off detection of an image.

    Raises
    ------
    TypeError
        If the sample is neither a string nor a PIL Image object.
    """

    image: Image.Image
    if isinstance(sample, str):
        image = Image.open(sample)
    elif isinstance(sample, Image.Image):
        image = sample
    else:
        raise TypeError("Unknown sample type")

    if device is None:
        device = torch.device("cpu")

    input_tensor = transform(image).unsqueeze(dim=0).to(device)
    detections = infer_batch(net, input_tensor, **kwargs)
    if score_threshold is not None:
        for i, detection in enumerate(detections):
            idxs = torch.where(detection["scores"] > score_threshold)
            detections[i]["scores"] = detection["scores"][idxs]
            detections[i]["boxes"] = detection["boxes"][idxs]
            detections[i]["labels"] = detection["labels"][idxs]

    detections = InferenceTransform.postprocess(
        detections, [input_tensor.shape[2:]], [image.size[::-1]]  # type: ignore[list-item]
    )

    return detections[0]


def infer_batch(
    net: torch.nn.Module | torch.ScriptModule,
    inputs: torch.Tensor,
    masks: Optional[torch.Tensor] = None,
    image_sizes: Optional[list[list[int]]] = None,
    **kwargs: Any,
) -> list[dict[str, torch.Tensor]]:
    (detections, _) = net(inputs, masks=masks, image_sizes=image_sizes, **kwargs)
    return detections  # type: ignore[no-any-return]


def infer_dataloader(
    device: torch.device,
    net: torch.nn.Module | torch.ScriptModule,
    dataloader: DataLoader,
    model_dtype: torch.dtype = torch.float32,
    amp: bool = False,
    amp_dtype: Optional[torch.dtype] = None,
    num_samples: Optional[int] = None,
    batch_callback: Optional[
        Callable[[list[str], torch.Tensor, list[dict[str, torch.Tensor]], list[dict[str, Any]], list[list[int]]], None]
    ] = None,
) -> tuple[list[str], list[dict[str, torch.Tensor]], list[dict[str, Any]]]:
    """
    Perform inference on a DataLoader using a given neural network.

    This function runs inference on a dataset provided through a DataLoader,
    optionally using mixed precision (amp).
    All returned detections and targets are transformed to original
    image coordinates regardless of the model's inference resolution.

    Parameters
    ----------
    device
        The device to run the inference on.
    net
        The model to use for inference.
    dataloader
        The DataLoader containing the dataset to perform inference on.
    model_dtype
        The base dtype to use.
    amp
        Whether to use automatic mixed precision.
    amp_dtype
        The mixed precision dtype.
    num_samples
        The total number of samples in the dataloader.
    batch_callback
        A function to be called after each batch is processed. If provided, it
        should accept four arguments:
        - list[str]: A list of file paths for the current batch
        - torch.Tensor: The input tensor for the current batch
        - list[dict[str, torch.Tensor]]: The detections for the current batch
        - list[dict[str, Any]]: A list of targets for the current batch
        - list[list[int]]: The image sizes for the current batch

    Returns
    -------
    A tuple containing three elements:
    - list[str]: A list of all processed file paths.
    - list[dict[str, torch.Tensor]]: A list of detection dictionaries.
    - list[dict[str, Any]]: A list of all targets.

    Notes
    -----
    - The function uses a progress bar (tqdm) to show the inference progress.
    - If 'num_samples' is not provided, the progress bar may not accurately
      reflect the total number of samples processed.
    - The batch_callback, if provided, is called after each batch is processed,
      allowing for real-time analysis or logging of results.
    """

    net.to(device, dtype=model_dtype)
    detections_list: list[dict[str, torch.Tensor]] = []
    target_list: list[dict[str, Any]] = []
    sample_paths: list[str] = []
    batch_size = dataloader.batch_size
    with tqdm(total=num_samples, initial=0, unit="images", unit_scale=True, leave=False) as progress:
        for file_paths, inputs, targets, orig_sizes, masks, image_sizes in dataloader:
            # Inference
            inputs = inputs.to(device, dtype=model_dtype, non_blocking=True)
            masks = masks.to(device, non_blocking=True)

            with torch.amp.autocast(device.type, enabled=amp, dtype=amp_dtype):
                detections = infer_batch(net, inputs, masks, image_sizes)

            detections = InferenceTransform.postprocess(detections, image_sizes, orig_sizes)
            if targets[0] != settings.NO_LABEL:
                targets = InferenceTransform.postprocess(targets, image_sizes, orig_sizes)

            detections_list.extend(detections)

            # Set targets and sample list
            target_list.extend(targets)
            sample_paths.extend(file_paths)

            if batch_callback is not None:
                batch_callback(file_paths, inputs, detections, targets, image_sizes)

            # Update progress bar
            progress.update(n=batch_size)

    return (sample_paths, detections_list, target_list)



from birder.net.detection.deformable_detr import Deformable_DETR
from birder.net.detection.detr import DETR
from birder.net.detection.efficientdet import EfficientDet
from birder.net.detection.faster_rcnn import Faster_RCNN
from birder.net.detection.fcos import FCOS
from birder.net.detection.retinanet import RetinaNet
from birder.net.detection.ssd import SSD
from birder.net.detection.ssdlite import SSDLite
from birder.net.detection.vitdet import ViTDet

__all__ = [
    "Deformable_DETR",
    "DETR",
    "EfficientDet",
    "Faster_RCNN",
    "FCOS",
    "RetinaNet",
    "SSD",
    "SSDLite",
    "ViTDet",
]



"""
UniFormer, adapted from
https://github.com/Sense-X/UniFormer/blob/main/image_classification/models/uniformer.py

Paper "UniFormer: Unifying Convolution and Self-attention for Visual Recognition", https://arxiv.org/abs/2201.09450
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.layers import LayerScale
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class MLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = nn.GELU()
        self.fc2 = nn.Linear(hidden_features, in_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class ConvMLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_features, in_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class Attention(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, attn_drop: float, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )
        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class ConvBlock(nn.Module):
    def __init__(self, dim: int, mlp_ratio: float, drop: float, drop_path: float) -> None:
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)
        self.norm1 = nn.BatchNorm2d(dim)
        self.conv1 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.conv2 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.attn = nn.Conv2d(dim, dim, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2), groups=dim)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.BatchNorm2d(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = ConvMLP(dim, hidden_features=mlp_hidden_dim, drop=drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.pos_embed(x)
        x = x + self.drop_path(self.conv2(self.attn(self.conv1(self.norm1(x)))))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class AttentionBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        layer_scale_init_value: Optional[float],
        drop: float,
        attn_drop: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)
        self.norm1 = nn.LayerNorm(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = MLP(dim, hidden_features=mlp_hidden_dim, drop=drop)

        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(dim, layer_scale_init_value)
            self.layer_scale_2 = LayerScale(dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.pos_embed(x)
        (B, N, H, W) = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = x + self.drop_path(self.layer_scale_1(self.attn(self.norm1(x))))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(self.norm2(x))))
        x = x.transpose(1, 2).reshape(B, N, H, W)

        return x


class PatchEmbed(nn.Module):
    def __init__(self, patch_size: tuple[int, int], in_channels: int, embed_dim: int) -> None:
        super().__init__()
        self.patch_size = patch_size
        self.norm = nn.LayerNorm(embed_dim)
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_size, padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        (B, _, H, W) = x.size()  # B, C, H, W
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        return x


class UniFormerStage(nn.Module):
    def __init__(
        self,
        block_type: Literal["conv", "attn"],
        patch_size: tuple[int, int],
        in_channels: int,
        embed_dim: int,
        depth: int,
        num_heads: int,
        mlp_ratio: float,
        layer_scale_init_value: Optional[float],
        drop: float,
        drop_path: list[float],
    ) -> None:
        super().__init__()
        self.patch_embed = PatchEmbed(patch_size=patch_size, in_channels=in_channels, embed_dim=embed_dim)

        layers = []
        for i in range(depth):
            if block_type == "conv":
                layers.append(ConvBlock(embed_dim, mlp_ratio=mlp_ratio, drop=drop, drop_path=drop_path[i]))
            elif block_type == "attn":
                layers.append(
                    AttentionBlock(
                        embed_dim,
                        num_heads=num_heads,
                        mlp_ratio=mlp_ratio,
                        qkv_bias=True,
                        layer_scale_init_value=layer_scale_init_value,
                        drop=0.0,
                        attn_drop=0.0,
                        drop_path=drop_path[i],
                    )
                )
            else:
                raise ValueError(f"Block type: {block_type} not supported")

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        x = self.blocks(x)

        return x


class UniFormer(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        block_type = ["conv", "conv", "attn", "attn"]
        patch_size = [4, 2, 2, 2]
        depth: list[int] = self.config["depth"]
        embed_dim: list[int] = self.config["embed_dim"]
        mlp_ratio: list[float] = self.config["mlp_ratio"]
        head_dim: int = self.config["head_dim"]
        drop_path_rate: float = self.config["drop_path_rate"]
        layer_scale_init_value: Optional[float] = self.config["layer_scale_init_value"]

        num_stages = len(depth)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depth)).split(depth)]
        num_heads = [dim // head_dim for dim in embed_dim]

        prev_dim = self.input_channels
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = UniFormerStage(
                block_type=block_type[i],  # type: ignore[arg-type]
                patch_size=(patch_size[i], patch_size[i]),
                in_channels=prev_dim,
                embed_dim=embed_dim[i],
                depth=depth[i],
                num_heads=num_heads[i],
                mlp_ratio=mlp_ratio[i],
                layer_scale_init_value=layer_scale_init_value,
                drop=0.0,
                drop_path=dpr[i],
            )
            return_channels.append(embed_dim[i])
            prev_dim = embed_dim[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dim[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "uniformer_s",
    UniFormer,
    config={
        "depth": [3, 4, 8, 3],
        "embed_dim": [64, 128, 320, 512],
        "mlp_ratio": [4.0, 4.0, 4.0, 4.0],
        "head_dim": 64,
        "drop_path_rate": 0.1,
        "layer_scale_init_value": None,
    },
)
registry.register_model_config(
    "uniformer_b",
    UniFormer,
    config={
        "depth": [5, 8, 20, 7],
        "embed_dim": [64, 128, 320, 512],
        "mlp_ratio": [4.0, 4.0, 4.0, 4.0],
        "head_dim": 64,
        "drop_path_rate": 0.3,
        "layer_scale_init_value": None,
    },
)
registry.register_model_config(
    "uniformer_l",
    UniFormer,
    config={
        "depth": [5, 10, 24, 7],
        "embed_dim": [128, 192, 448, 640],
        "mlp_ratio": [4.0, 4.0, 4.0, 4.0],
        "head_dim": 64,
        "drop_path_rate": 0.4,
        "layer_scale_init_value": 1e-6,
    },
)
